"""Test the readers."""

import datetime
import os
import sys
import unittest
from unittest import mock

import numpy as np
import numpy.testing
import pytest
import xarray as xr

from pygac.gac_pod import scanline
from pygac.gac_reader import GACReader, ReaderError
from pygac.lac_pod import LACPODReader
from pygac.lac_pod import scanline as lacpod_scanline
from pygac.lac_reader import LACReader
from pygac.pod_reader import POD_QualityIndicator, header3
from pygac.pod_reader import tbm_header as tbm_header_dtype
from pygac.reader import NoTLEData


class FakePath(os.PathLike):
    """Fake path class."""

    def __init__(self, path):
        """Initialize the path."""
        self.path = str(path)

    def __fspath__(self):
        """Return the path."""
        return self.path


class FakeGACReader(GACReader):
    """Fake GAC reader class."""

    QFlag = POD_QualityIndicator
    _quality_indicators_key = "quality_indicators"
    tsm_affected_intervals = {None: []}
    along_track = 3
    across_track = 409

    def __init__(self, *args, **kwargs):
        """Initialize the fake reader."""
        super().__init__(*args, **kwargs)
        self.scan_width = self.across_track
        scans = np.zeros(self.along_track, dtype=scanline)
        scans["scan_line_number"] = np.arange(self.along_track)
        scans["sensor_data"] = 128
        self.scans = scans
        self.head = {"foo": "bar"}
        self.head = np.rec.fromrecords([("bar", ),],names="foo")
        self.spacecraft_name = "noaa6"

    def _get_times_from_file(self):
        year = np.full(self.along_track, 1970, dtype=int)
        jday = np.full(self.along_track, 1, dtype=int)
        msec = 1000 * np.arange(1, self.along_track+1, dtype=int)
        return year, jday, msec

    def get_header_timestamp(self):
        """Get the header timestamp."""
        return datetime.datetime(1970, 1, 1)

    def get_telemetry(self):
        """Get the telemetry."""
        prt = 51 * np.ones(self.along_track)  # prt threshold is 50
        prt[::5] = 0
        ict = 101 * np.ones((self.along_track, 3))  # ict threshold is 100
        space = 101 * np.ones((self.along_track, 3))  # space threshold is 100
        return prt, ict, space

    def _adjust_clock_drift(self):
        pass

    def _get_lonlat_from_file(self):
        return np.zeros((self.along_track, 51)), np.zeros((self.along_track, 51))

    @staticmethod
    def _get_ir_channels_to_calibrate():
        return [3, 4, 5]

    def postproc(self, channels):
        """Postprocess the data."""
        pass

    def read(self, filename, fileobj=None):
        """Read the data."""
        pass

    @classmethod
    def read_header(cls, filename, fileobj=None):
        """Read the header."""
        pass

    def get_tsm_pixels(self, channels):
        """Get the tsm pixels."""
        pass

    def get_tle_lines(self):
        """Get tle lines."""
        pass

class FakeGACReaderWithWrongPRTs(FakeGACReader):

    along_track = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scans["scan_line_number"] = np.array([1, 3, 4, 5, 6, 7, 8, 9, 10, 11])


class TestGacReader(unittest.TestCase):
    """Test the common GAC Reader."""

    longMessage = True

    @mock.patch.multiple("pygac.gac_reader.GACReader",
                         __abstractmethods__=set())
    def setUp(self, *mocks):
        """Set up the tests."""
        self.reader = GACReader()

    def test_filename(self):
        """Test the setter of the filename property."""
        # test path with .gz extension
        filename = "NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"
        filepath = "/path/to/" + filename + ".gz"
        self.reader.filename = filepath
        self.assertEqual(self.reader.filename, filename)
        self.reader.filename = None
        self.assertIsNone(self.reader.filename)
        self.reader.filename = FakePath(filepath)
        self.assertEqual(self.reader.filename, filename)

    @unittest.skipIf(sys.version_info.major < 3, "Skipped in python2!")
    def test__read_scanlines(self):
        """Test the scanline extraction."""
        self.reader.scanline_type = np.dtype([
            ("a", "S2"), ("b", "<i4")])
        # request more scan lines than available
        with self.assertWarnsRegex(RuntimeWarning,
                                   "Unexpected number of scanlines!"):
            buffer = (b"a\x00\x01\x00\x00\x00"
                      b"b\x00\x02\x00\x00\x00"
                      b"c\x00\x03\x00\x00\x00")
            count = 4
            self.reader._read_scanlines(buffer, count)
        # check the output
        first_line = self.reader.scans[0]
        self.assertEqual(first_line["a"], b"a")
        self.assertEqual(first_line["b"], 1)

    def test__validate_header(self):
        """Test the header validation."""
        # wrong name pattern
        with self.assertRaisesRegex(ReaderError,
                                    "Data set name .* does not match!"):
            head = {"data_set_name": b"abc.txt"}
            self.reader._validate_header(head)
        # Unicode errors are now caught with the same exception.
        with self.assertRaisesRegex(ReaderError,
                                    "Data set name .* does not match!"):
            head = {"data_set_name": b"\xea\xf8"}
            self.reader._validate_header(head)

    def test__correct_data_set_name_ebcdic_encoded_header_invalid_path(self):
        """Test the data_set_name correction in file header."""
        inv_filename = "InvalidFileName"
        inv_filepath = "path/to/" + inv_filename

        expected_data_set_name = "NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"
        val_head = {"data_set_name": "NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI".encode("cp500")}
        head = self.reader._correct_data_set_name(val_head.copy(), inv_filepath)
        assert head["data_set_name"] == expected_data_set_name.encode()

    def test__correct_data_set_name_valid_header_and_file(self):
        """Test the data_set_name correction in file header."""
        val_filename = "NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"
        val_filepath = "path/to/" + val_filename
        val_head = {"data_set_name": b"NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"}
        # Note: always pass a copy to _correct_data_set_name, because
        #       the input header is modified in place.
        # enter a valid data_set_name and filepath
        head = self.reader._correct_data_set_name(val_head.copy(), val_filepath)
        assert head["data_set_name"] == val_filename.encode()

    def test__correct_data_set_name_invalid_header_and_valid_file(self):
        """Test the data_set_name correction in file header."""
        val_filename = "NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"
        val_filepath = "path/to/" + val_filename
        inv_head = {"data_set_name": b"InvalidDataSetName"}

        # enter an invalid data_set_name, but valid filepath
        head = self.reader._correct_data_set_name(inv_head.copy(), val_filepath)
        assert head["data_set_name"] == val_filename.encode()

    def test__correct_data_set_name_invalid_header_and_file(self):
        """Test the data_set_name correction in file header."""
        inv_filename = "InvalidFileName"
        inv_filepath = "path/to/" + inv_filename
        inv_head = {"data_set_name": b"InvalidDataSetName"}
        with self.assertRaisesRegex(ReaderError, "Cannot determine data_set_name!"):
            _ = self.reader._correct_data_set_name(inv_head.copy(), inv_filepath)

    def test__correct_data_set_name_valid_header_invalid_file(self):
        """Test the data_set_name correction in file header."""
        val_head = {"data_set_name": b"NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"}
        inv_filename = "InvalidFileName"
        inv_filepath = "path/to/" + inv_filename

        # enter a valid data_set_name, and an invalid filepath
        # should be fine, because the data_set_name is the pefered source
        head = self.reader._correct_data_set_name(val_head.copy(), inv_filepath)
        assert head["data_set_name"] == val_head["data_set_name"]

    def test__correct_data_set_name_valid_header_pathlib_file(self):
        """Test the data_set_name correction in file header."""
        val_filename = "NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"
        val_filepath = "path/to/" + val_filename
        val_head = {"data_set_name": b"NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"}

        fs_filepath = FakePath(val_filepath)
        head = self.reader._correct_data_set_name(val_head.copy(), fs_filepath)
        self.assertEqual(head["data_set_name"], val_filename.encode())

    @mock.patch("pygac.reader.Reader.get_calibrated_channels")
    def test__get_calibrated_channels_uniform_shape(self, get_channels):
        """Test the uniform shape as required by gac_io.save_gac."""
        # check if it raises the assertion error
        channels = np.arange(2*2*5, dtype=float).reshape((2, 2, 5))
        get_channels.return_value = channels
        with self.assertRaises(AssertionError):
            self.reader._get_calibrated_channels_uniform_shape()

    def test_get_calibrated_channels(self):
        """Test getting calibrated channels."""
        reader = FakeGACReader()
        res = reader.get_calibrated_channels()

        np.testing.assert_allclose(res[:, 1, 0], 8.84714652)
        np.testing.assert_allclose(res[:, 2, 1], 10.23511303)
        assert reader.meta_data["calib_coeffs_version"] == "PATMOS-x, v2023"

    def test_get_calibrated_channels_with_wrong_prts(self):
        """Test getting calibrated channels."""
        reader = FakeGACReaderWithWrongPRTs()
        res = reader.get_calibrated_channels()

        np.testing.assert_allclose(res[:, 1, 0], 8.84714652)
        np.testing.assert_allclose(res[:, 2, 1], 10.23511303)
        assert reader.meta_data["calib_coeffs_version"] == "PATMOS-x, v2023"

    def test_to_datetime64(self):
        """Test conversion from (year, jday, msec) to datetime64."""
        t0 = GACReader.to_datetime64(year=np.array(1970), jday=np.array(1),
                                     msec=np.array(0))
        self.assertEqual(t0.astype("i8"), 0,
                         msg="Conversion (year, jday, msec) to datetime64 "
                             "is not correct")

    def test_to_datetime(self):
        """Test conversion from datetime64 to datetime."""
        dt = datetime.datetime(2019, 10, 1, 12, 15, 18, 12345)
        dt64 = np.datetime64(dt)
        self.assertEqual(self.reader.to_datetime(dt64), dt)

    def test_lineno2msec(self):
        """Test scanline timestamp estimation."""
        self.assertEqual(self.reader.lineno2msec(12345), 6172000)

    @mock.patch("pygac.reader.Reader.update_meta_data")
    @mock.patch("pygac.gac_reader.GACReader._get_lonlat_from_file")
    @mock.patch("pygac.gac_reader.GACReader._get_corrupt_mask")
    @mock.patch("pygac.gac_reader.GACReader._adjust_clock_drift")
    @mock.patch("pygac.reader.Reader.lonlat_interpolator")
    def test_get_lonlat(self, lonlat_interpolator, adjust_clockdrift,
                        get_corrupt_mask, get_lonlat,
                        update_meta_data):
        """Test common lon/lat computation."""
        lon_i = np.array([np.nan, 1, 2, 3, -180.1, 180.1])
        lat_i = np.array([1, 2, 3, np.nan, -90.1, 90.1])
        get_lonlat.return_value = lon_i, lat_i
        lonlat_interpolator.return_value = lon_i, lat_i
        get_corrupt_mask.return_value = np.array(
            [0, 0, 1, 0, 0, 0], dtype=bool)

        lons_exp = np.array([np.nan, 1, np.nan, 3., np.nan, np.nan])
        lats_exp = np.array([1, 2, np.nan, np.nan, np.nan, np.nan])

        # Default
        lons, lats = self.reader.get_lonlat()
        get_lonlat.assert_called()
        update_meta_data.assert_called()
        adjust_clockdrift.assert_called()
        numpy.testing.assert_array_equal(lons, lons_exp)
        numpy.testing.assert_array_equal(lats, lats_exp)

        # Interpolation disabled
        lonlat_interpolator.reset_mock()
        adjust_clockdrift.reset_mock()
        self.reader.clock_drift_correction_applied = False
        self.reader.interpolate_coords = False
        self.reader.adjust_clock_drift = True
        self.reader.lons = self.reader.lats = None
        self.reader.get_lonlat()
        numpy.testing.assert_array_equal(lons, lons_exp)
        numpy.testing.assert_array_equal(lats, lats_exp)
        lonlat_interpolator.assert_not_called()
        adjust_clockdrift.assert_called()

        # Clock drift adjustment disabled
        lonlat_interpolator.reset_mock()
        adjust_clockdrift.reset_mock()
        self.reader.interpolate_coords = True
        self.reader.adjust_clock_drift = False
        self.reader.lons = self.reader.lats = None
        self.reader.get_lonlat()
        numpy.testing.assert_array_equal(lons, lons_exp)
        numpy.testing.assert_array_equal(lats, lats_exp)
        lonlat_interpolator.assert_called()
        adjust_clockdrift.assert_not_called()

        # Test caching
        methods = [get_lonlat, lonlat_interpolator,
                   adjust_clockdrift, get_corrupt_mask]
        for method in methods:
            method.reset_mock()
        self.reader.get_lonlat()
        for method in methods:
            method.asser_not_called()

    @mock.patch("pygac.reader.Reader.update_meta_data")
    @mock.patch("pygac.gac_reader.GACReader._get_corrupt_mask")
    @mock.patch("pygac.gac_reader.GACReader._adjust_clock_drift")
    @mock.patch("pygac.gac_reader.GACReader._get_lonlat_from_file")
    @mock.patch("pygac.reader.Reader.lonlat_interpolator")
    def test_interpolate(self, lonlat_interpolator, _get_lonlat, _adjust_clock_drift,
                         _get_corrupt_mask, update_meta_data):
        """Test interpolate method in get_lonlat."""
        self.lons = None
        self.lats = None
        rng = np.random.default_rng()
        lr_lons = 90 * rng.random((17, 51))
        lr_lats = 90 * rng.random((17, 51))
        _get_lonlat.return_value = lr_lons, lr_lats
        self.interpolate_coords = True
        lonlat_interpolator.reset_mock()
        lonlat_interpolator.return_value = (90 * rng.random((17, 409)),
                                          90 * rng.random((17, 409)))
        lons, lats = self.reader.get_lonlat()
        self.assertEqual(lons.shape[1], 409)
        lonlat_interpolator.assert_called_once_with(lr_lons, lr_lats)

    @mock.patch("pygac.gac_reader.GACReader._get_corrupt_mask")
    def test_get_corrupt_mask(self, get_corrupt_mask):
        """Test common computation of corrupt scanline mask."""
        get_corrupt_mask.return_value = [1, 2, 3]
        self.assertEqual(self.reader.mask, [1, 2, 3])

        # Test caching
        get_corrupt_mask.reset_mock()
        self.reader.mask
        get_corrupt_mask.assert_not_called()

    def test_midnight_scanline(self):
        """Test midnight scanline computation."""
        # Define test cases...
        # ... midnight scanline exists
        utcs1 = np.array([-3, -2, -1, 0, 1, 2, 3]).astype("datetime64[ms]")
        scanline1 = 2

        # ... midnight scanline does not exist
        utcs2 = np.array([1, 2, 3]).astype("datetime64[ms]")
        scanline2 = None

        for utcs, scan_line in zip((utcs1, utcs2), (scanline1, scanline2)):
            self.reader._times_as_np_datetime64 = utcs
            self.assertEqual(self.reader.get_midnight_scanline(), scan_line,
                             msg="Incorrect midnight scanline")

    def test_miss_lines(self):
        """Test detection of missing scanlines."""
        lines = [2, 4, 5, 6, 10, 11, 12]
        miss_lines_ref = [1, 3, 7, 8, 9]
        self.reader.scans = np.zeros(
            len(lines), dtype=[("scan_line_number", "i2")])
        self.reader.scans["scan_line_number"] = lines
        miss_lines = self.reader.get_miss_lines()
        self.assertTrue((miss_lines == miss_lines_ref).all(),
                        msg="Missing scanlines not detected correctly")
        self.assertEqual(miss_lines.dtype, int)

    def test_tle2datetime64(self, *mocks):
        """Test conversion from TLE timestamps to datetime64."""
        dates = np.array([70365.1234, 18001.25])
        dates64_exp = [np.datetime64(datetime.datetime(1970, 12, 31, 2, 57, 41, 760000), "[ms]"),
                       np.datetime64(datetime.datetime(2018, 1, 1, 6, 0), "[ms]")]
        dates64 = GACReader.tle2datetime64(dates)
        self.assertTrue(np.all(dates64 == dates64_exp))

    @mock.patch("pygac.gac_reader.GACReader.get_times")
    @mock.patch("pygac.gac_reader.GACReader.get_tle_file")
    @mock.patch("pygac.gac_reader.GACReader.read_tle_file")
    def test_get_tle_lines(self, read_tle_file, *mocks):
        """Test identification of closest TLE lines."""
        tle_data = ["1 38771U 12049A   18363.63219793 -.00000013  00000-0  14176-4 0  9991\r\n",
                    "2 38771  98.7297  60.1350 0002062  95.9284  25.0713 14.21477560325906\r\n",
                    "1 38771U 12049A   18364.62426010 -.00000015  00000-0  13136-4 0  9990\r\n",  # 2018-12-30 14:58
                    "2 38771  98.7295  61.1159 0002062  94.5796  60.2561 14.21477636326047\r\n",
                    "1 38771U 12049A   18364.94649306 -.00000018  00000-0  12040-4 0  9996\r\n",  # 2018-12-30 22:42
                    "2 38771  98.7295  61.4345 0002060  94.1226 268.7521 14.21477633326092\r\n",
                    "1 38771U 12049A   18365.81382142 -.00000015  00000-0  13273-4 0  9991\r\n",
                    "2 38771  98.7294  62.2921 0002057  92.7653  26.0030 14.21477711326215\r\n"]

        expected = {
            datetime.datetime(2018, 12, 20, 12, 0): None,
            datetime.datetime(2018, 12, 28, 12, 0): 0,
            datetime.datetime(2018, 12, 30, 16, 0): 2,
            datetime.datetime(2018, 12, 30, 20, 0): 4,
            datetime.datetime(2019, 1, 1, 12, 0): 6,
            datetime.datetime(2019, 1, 8, 12, 0): None
        }

        read_tle_file.return_value = tle_data
        for time, tle_idx in expected.items():
            self.reader._times_as_np_datetime64 = np.array([time], dtype="datetime64[ms]")
            self.reader.tle_lines = None
            if tle_idx is None:
                self.assertRaises(NoTLEData, self.reader.get_tle_lines)
            else:
                tle1, tle2 = self.reader.get_tle_lines()
                self.assertEqual(tle1, tle_data[tle_idx])
                self.assertEqual(tle2, tle_data[tle_idx + 1])

    @mock.patch("pygac.reader.Reader.get_tle_lines")
    def test_get_sat_angles_without_tle_at_nadir(self, get_tle_lines):
        """Test that the get satellite angles without tle at nadir."""
        get_tle_lines.side_effect = NoTLEData("No TLE data available")
        rng = np.random.RandomState(125)
        self.reader.lons = rng.rand(100, 409) * 90
        self.reader.lats = rng.rand(100, 409) * 90
        self.reader._times_as_np_datetime64 = np.array(
            [numpy.datetime64(datetime.datetime(1980, 1, 3, 11, 47, 15, 469000)) for date in range(100)])
        sat_azi, sat_elev = self.reader.get_sat_angles()
        self.assertEqual(np.sum(np.isnan(sat_elev)), 0)
        np.testing.assert_allclose(sat_elev[:, 204], 90., atol=0.01)

    @mock.patch("pygac.reader.Reader.get_tle_lines")
    def test_get_sat_angles_without_tle(self, get_tle_lines):
        """Test the get satellite angles without tle."""
        get_tle_lines.side_effect = NoTLEData("No TLE data available")

        # Test data correspond to columns 0:2, 201:208 and 407:409. Extracted like this:
        # self.lons[0:5, [0, 1, 201, 202, 203, 204, 205, 206, 207, -2, -1]]
        self.reader.lons = np.array([[69.41555135, 68.76815744, 28.04133742, 27.94671757, 27.85220562,
                                      27.7578125, 27.66354783, 27.5694182, 27.47542957, 2.66416611,
                                      2.39739436],
                                     [69.41409536, 68.76979616, 28.00228658, 27.9076628, 27.8131467,
                                      27.71875, 27.62448295, 27.53035209, 27.43636312, 2.61727408,
                                      2.35049275],
                                     [69.42987929, 68.78543423, 27.96407251, 27.86936406, 27.77457923,
                                      27.6796875, 27.58467527, 27.48959853, 27.39453053, 2.5704025,
                                      2.30362323],
                                     [69.44430772, 68.80064104, 27.91910034, 27.82340242, 27.72794715,
                                      27.6328125, 27.53805662, 27.44366144, 27.34959008, 2.53088093,
                                      2.26729486],
                                     [69.47408815, 68.8259859, 27.87666513, 27.78224611, 27.68795045,
                                      27.59375, 27.49962326, 27.40557682, 27.31162435, 2.48359319,
                                      2.21976689]])
        self.reader.lats = np.array([[71.62830288, 71.67081539, 69.90976034, 69.89297223, 69.87616536,
                                      69.859375, 69.84262997, 69.82593315, 69.80928089, 61.61334632,
                                      61.466097],
                                     [71.65222644, 71.69455292, 69.9331981, 69.91640991, 69.89960295,
                                      69.8828125, 69.86606741, 69.84937054, 69.83271828, 61.62903602,
                                      61.48182119],
                                     [71.68324681, 71.7256586, 69.96444365, 69.94765637, 69.93085095,
                                      69.9140625, 69.89731955, 69.8806245, 69.86397337, 61.65247459,
                                      61.50526043],
                                     [71.71489701, 71.75715746, 69.98789144, 69.97110185, 69.9542928,
                                      69.9375, 69.92075278, 69.90405431, 69.88740114, 61.66489162,
                                      61.51597801],
                                     [71.7390608, 71.78108857, 70.01126173, 69.99449392, 69.97770882,
                                      69.9609375, 69.9442054, 69.92751555, 69.91086544, 61.67769951,
                                      61.52802445]])
        self.reader._times_as_np_datetime64 = np.array(
            [numpy.datetime64(datetime.datetime(1980, 1, 3, 11, 47, 15, 469000)),
             numpy.datetime64(datetime.datetime(1980, 1, 3, 11, 47, 15, 969000)),
             numpy.datetime64(datetime.datetime(1980, 1, 3, 11, 47, 16, 469000)),
             numpy.datetime64(datetime.datetime(1980, 1, 3, 11, 47, 16, 969000)),
             numpy.datetime64(datetime.datetime(1980, 1, 3, 11, 47, 17, 469000))])
        expected_sat_azi_0 = np.array([283.09872924, 283.12775589, 283.13951497, 283.14786413, 283.19638805])
        expected_sat_azi_201 = np.array([272.85051989, 273.79847634, 272.04794616, 273.01363377, 274.00055397])
        expected_sat_azi_408 = np.array([39.77021472, 39.71516966, 39.68104134, 39.60503726, 39.5403431])
        expected_sat_elev_0 = np.array([20.94889204, 20.96041284, 20.96368521, 20.96826309, 20.95849212])
        expected_sat_elev_204 = np.array([89.24533884, 89.22663677, 89.25079817, 89.24938043, 89.23004118])
        sat_azi, sat_elev = self.reader.get_sat_angles()
        np.testing.assert_allclose(sat_azi[:, 0], expected_sat_azi_0, atol=1.0)
        np.testing.assert_allclose(sat_azi[:, 2], expected_sat_azi_201, atol=35.0)  # Azi bad close to center!
        np.testing.assert_allclose(sat_azi[:, -1], expected_sat_azi_408, atol=1.0)
        np.testing.assert_allclose(sat_elev[:, 0], expected_sat_elev_0, atol=1.0)
        np.testing.assert_allclose(sat_elev[:, 5], expected_sat_elev_204, atol=1.0)

    @mock.patch("pygac.gac_reader.GACReader._get_corrupt_mask")
    def test_get_angles(self, get_corrupt_mask):
        """Test get_angles function of the reader."""
        # Line: 1, 649, 6198 and 12658 from Tiros-N file (1980-01-03 11:47)
        lon_i = np.array(
            [69.41555, 152.10587, 164.3131, 67.23855, np.nan])[:, np.newaxis]
        lat_i = np.array(
            [71.6283, 85.24265, -62.076958, 82.72296, np.nan])[:, np.newaxis]
        get_corrupt_mask.return_value = np.isnan(lon_i)
        self.reader.lons = lon_i
        self.reader.lats = lat_i
        self.reader.tle_lines = [
            '1 11060U 78096A   80003.54792075  .00000937  00000-0  52481-3 0  2588\r\n',  # noqa
            '2 11060  98.9783 332.1605 0012789  88.8047 271.4583 14.11682873 63073\r\n']  # noqa
        self.reader._times_as_np_datetime64 = np.array(
            [315748035469, 315748359969,
             315751135469, 315754371969,
             315754371969]).astype("datetime64[ms]")
        self.reader.spacecrafts_orbital = {25: "tiros n"}
        self.reader.spacecraft_id = 25
        expected_sat_azi = np.array(
            [-76.90, 11.08, 145.33, -50.01, np.nan])[:, np.newaxis]
        expected_sun_azi = np.array(
            [-120.36, -31.94, -173.51, -93.67, np.nan])[:, np.newaxis]
        expected_sat_zenith = np.array(
            [69.05, 69.04, 69.55, 69.07, np.nan])[:, np.newaxis]
        expected_sun_zenith = np.array(
            [104.30, 116.94, 94.86, 112.60, np.nan])[:, np.newaxis]
        expected_rel_azi = np.array(
            [43.45, 43.01, 41.16, 43.65, np.nan])[:, np.newaxis]

        retv = self.reader.get_angles()
        (sat_azi, sat_zenith, sun_azi, sun_zenith, rel_azi) = retv
        np.testing.assert_allclose(sat_azi, expected_sat_azi, atol=0.01)
        np.testing.assert_allclose(sun_azi, expected_sun_azi, atol=0.01)
        np.testing.assert_allclose(sat_zenith, expected_sat_zenith, atol=0.01)
        np.testing.assert_allclose(sun_zenith, expected_sun_zenith, atol=0.01)
        np.testing.assert_allclose(rel_azi, expected_rel_azi, atol=0.01)

    def test_get_tle_file(self):
        """Test get_tle_file."""
        self.reader.tle_dir = "/tle/dir"
        self.reader.tle_name = "tle_%(satname)s.txt"
        self.reader.spacecraft_name = "ISS"
        tle_file = self.reader.get_tle_file()
        self.assertEqual(tle_file, "/tle/dir/tle_ISS.txt")

    @mock.patch("pygac.gac_reader.GACReader.get_tsm_pixels")
    def test_mask_tsm_pixels(self, get_tsm_pixels):
        """Test masking of pixels affected by the scan motor issue."""
        get_tsm_pixels.return_value = ([0, 1], [0, 1])
        channels = np.array([[[1., 2., 3.],
                              [4., 5., 6.]],
                             [[7., 8., 9.],
                              [10., 11., 12.]]])  # (lines, pixels, channels)
        masked_exp = np.array([[[np.nan, np.nan, np.nan],
                                [4., 5., 6.]],
                               [[7., 8., 9.],
                                [np.nan, np.nan, np.nan]]])

        channels = xr.DataArray(channels, dims=["scan_line_index", "columns", "channel_name"],
                                coords=dict(channel_name=["1", "2", "3a"] ))
        ds = xr.Dataset(dict(channels=channels))

        self.reader.mask_tsm_pixels(ds)  # masks in-place
        numpy.testing.assert_array_equal(ds["channels"].values, masked_exp)

    def test_correct_scan_line_numbers(self):
        """Test scanline number correction."""
        scans, expected = _get_scanline_numbers(14000)
        self.reader.scans = scans
        self.reader.correct_scan_line_numbers()
        numpy.testing.assert_array_equal(self.reader.scans["scan_line_number"],
                                         expected)

    @mock.patch("pygac.gac_reader.GACReader.get_header_timestamp")
    def test_correct_times_thresh(self, get_header_timestamp):
        """Test correction of scanline timestamps."""
        header_time = datetime.datetime(2016, 8, 16, 16, 7, 36)

        # Create artificial timestamps
        _, scan_line_numbers = _get_scanline_numbers(14000)
        t0 = np.array([header_time], dtype="datetime64[ms]").astype("i8")[0]
        shift = 1000
        msecs = t0 + shift + scan_line_numbers / GACReader.scan_freq
        utcs_expected = msecs.copy().astype(">M8[ms]")

        # Add some corruptions
        msecs[3000:] += 1800 * 1000
        msecs[1000] += 24*3600*1000
        msecs[2000] -= 24*3600*1000

        # Mock reader
        get_header_timestamp.return_value = header_time
        self.reader._times_as_np_datetime64 = msecs.astype(">M8[ms]")
        self.reader.scans = np.array(scan_line_numbers,
                                     dtype=[("scan_line_number", ">u2")])

        # Test correction
        self.reader.correct_times_thresh()
        numpy.testing.assert_array_equal(self.reader._times_as_np_datetime64, utcs_expected)

    def test_calculate_sun_earth_distance_correction(self):
        """Test the calculate sun earth distance correction method."""
        self.reader._times_as_np_datetime64 = np.array([315748035469, 315748359969,
                                     315751135469, 315754371969,
                                     315754371969]).astype("datetime64[ms]")
        corr = self.reader.get_sun_earth_distance_correction()
        numpy.testing.assert_almost_equal(corr, 0.96660494, decimal=7)

    @mock.patch("pygac.reader.Reader.get_sun_earth_distance_correction")
    @mock.patch("pygac.reader.Reader.get_midnight_scanline")
    @mock.patch("pygac.reader.Reader.get_miss_lines")
    def test_update_metadata(self,
                             get_miss_lines,
                             get_midnight_scanline,
                             get_sun_earth_distance_correction):
        """Test updating the metadata."""
        get_miss_lines.return_value = "miss_lines"
        get_midnight_scanline.return_value = "midn_line"
        get_sun_earth_distance_correction.return_value = "factor"
        self.reader.head = {"foo": "bar"}

        self.reader.update_meta_data()

        mda_exp = {"midnight_scanline": "midn_line",
                   "missing_scanlines": "miss_lines",
                   "sun_earth_distance_correction_factor": "factor",
                   "gac_header": {"foo": "bar"}}
        self.assertDictEqual(self.reader.meta_data, mda_exp)


def _get_scanline_numbers(scanlines_along_track):
    """Create artificial scanline numbers with some corruptions.

    Returns:
        Corrupted and corrected scanline numbers.

    """
    scans = np.zeros(scanlines_along_track, dtype=[("scan_line_number", ">u2")])
    scans["scan_line_number"] = np.arange(1, scanlines_along_track + 1)

    # ... with 500 missing scanlines at scanline 8000
    scans["scan_line_number"][8000:] += 500
    corrected = scans["scan_line_number"].copy()

    # ... and some spikes here and there
    scans["scan_line_number"][3000] += 1E4
    scans["scan_line_number"][9000] -= 1E4
    corrected = np.delete(corrected, [3000, 9000])

    return scans, corrected


class TestLacReader(unittest.TestCase):
    """Test the common LAC Reader."""

    longMessage = True

    @mock.patch.multiple("pygac.lac_reader.LACReader",
                         __abstractmethods__=set())
    def setUp(self, *mocks):
        """Set up the tests."""
        self.reader = LACReader()

    def test_lac_reader_accepts_FRAC(self):
        """Test the header validation."""
        head = {"data_set_name": b"NSS.FRAC.M1.D19115.S2352.E0050.B3425758.SV"}
        self.reader._validate_header(head)

    def test_correct_scan_line_numbers(self):
        """Test scanline number correction."""
        scans, expected = _get_scanline_numbers(22000)
        self.reader.scans = scans
        self.reader.correct_scan_line_numbers()
        numpy.testing.assert_array_equal(self.reader.scans["scan_line_number"],
                                         expected)


@pytest.fixture
def pod_file_with_tbm_header(tmp_path):
    """Create a pod file with a tbm header, and header, and some scanlines."""
    number_of_scans = 3

    tbm_header = np.zeros(1, dtype=tbm_header_dtype)
    tbm_header["data_set_name"] = "BRN.HRPT.NJ.D00322.S0334.E0319.B3031919.BL\x80\x80".encode("cp500")
    tbm_header["select_flag"] = b"S"
    tbm_header["beginning_latitude"] = b"+77"
    tbm_header["ending_latitude"] = b"+22"
    tbm_header["beginning_longitude"] = b"-004"
    tbm_header["ending_longitude"] = b"+032"
    tbm_header["start_hour"] = b"AL"
    tbm_header["start_minute"] = b"L "
    tbm_header["number_of_minutes"] = b"ALL"
    tbm_header["appended_data_flag"] = b"Y"
    tbm_header["channel_select_flag"][0, :5] = b"\x01"
    tbm_header["sensor_data_word_size"] = b"10"

    header = np.zeros(1, dtype=header3)
    header["noaa_spacecraft_identification_code"] = 3
    header["data_type_code"] = 48
    header["start_time"] = [51522, 181, 62790]
    header["number_of_scans"] = number_of_scans
    header["end_time"] = [51522, 195, 42286]
    header["processing_block_id"] = b"3031919"
    header["ramp_auto_calibration"] = 0
    header["number_of_data_gaps"] = 0
    header["dacs_quality"] = [21, 7, 0, 0, 0, 0]
    header["calibration_parameter_id"] = 12336
    header["dacs_status"] = 24
    header["nadir_earth_location_tolerance"] = 20
    header["start_of_data_set_year"] = 2000
    # EBCDIC, aka cp500 encoding
    header["data_set_name"] = (b"\xc2\xd9\xd5K\xc8\xd9\xd7\xe3K\xd5\xd1K\xc4\xf0\xf0\xf3\xf2\xf2K\xe2\xf0\xf3\xf1\xf9K"
                               b"\xc5\xf0\xf3\xf3\xf4K\xc2\xf3\xf0\xf3\xf1\xf9\xf1\xf9K\xc2\xd3@@")
    header["year_of_epoch"] = 2000
    header["julian_day_of_epoch"] = 322
    header["millisecond_utc_epoch_time_of_day"] = 11864806
    header["semi_major_axis"] = 7226239
    header["eccentricity"] = 100496
    header["inclination"] = 9915900
    header["argument_of_perigee"] = 2511798
    header["right_ascension"] = 30366227
    header["mean_anomaly"] = 7341437
    header["x_component_of_position_vector"] = 40028760
    header["y_component_of_position_vector"] = -60151283
    header["z_component_of_position_vector"] = 0,
    header["x_dot_component_of_position_vector"] = -990271
    header["y_dot_component_of_position_vector"] = -646066
    header["z_dot_component_of_position_vector"] = 7337861
    header["yaw_fixed_error_correction"] = 0
    header["roll_fixed_error_correction"] = 0
    header["pitch_fixed_error_correction"] = 0

    scanlines = np.zeros(number_of_scans, dtype=lacpod_scanline)
    scanlines["scan_line_number"] = np.arange(number_of_scans) + 1
    scanlines["time_code"][:, :2] = [51522, 181]
    scanlines["time_code"][:, 2] = np.arange(62790, 62790 + 166 * number_of_scans, 166)
    scanlines["quality_indicators"] = 1073741824
    scanlines["calibration_coefficients"] = [149722896, -23983736, 173651248, -27815402, -1803673, 6985664, -176321840,
                                             666680320, -196992576, 751880640]
    scanlines["number_of_meaningful_zenith_angles_and_earth_location_appended"] = 51

    one_line_angles = np.array([-24, -26, -28, -30, -32, -33, -34, -35, -37, -37, -38, -39, -40, -41, -41, -42,
                                -43, -43, -44, -45, -45, -46, -46, -47, -48, -48, -49, -49, -50, -50, -51, -52,
                                -52, -53, -53, -54, -55, -55, -56, -57, -58, -59, -60, -61, -62, -63, -64, -66,
                                -68, -70, -73])
    scanlines["solar_zenith_angles"] = [one_line_angles, one_line_angles + 1, one_line_angles + 2]

    one_line_lats = np.array([9929, 9966, 9983, 9988, 9984, 9975, 9963, 9948, 9931, 9913, 9895, 9875,
                              9856, 9836, 9816, 9796, 9775, 9755, 9734, 9713, 9692, 9671, 9650, 9628,
                              9606, 9583, 9560, 9537, 9513, 9488, 9463, 9437, 9409, 9381, 9351, 9320,
                              9288, 9253, 9216, 9177, 9135, 9090, 9040, 8986, 8926, 8859, 8784, 8697,
                              8596, 8476, 8327])

    scanlines["earth_location"]["lats"] = [one_line_lats - 1, one_line_lats, one_line_lats + 1]

    one_line_lons = np.array([-36, 747, 1415, 1991, 2492, 2931, 3320, 3667, 3979, 4262, 4519, 4755, 4972,
                              5174, 5362, 5538, 5703, 5860, 6009, 6150, 6286, 6416, 6542, 6663, 6781,
                              6896, 7009, 7119, 7227, 7334, 7440, 7545, 7650, 7755, 7860, 7966, 8073,
                              8181, 8291, 8403, 8518, 8637, 8759, 8886, 9019, 9159, 9307, 9466, 9637,
                              9824, 10033])

    scanlines["earth_location"]["lons"] = [one_line_lons, one_line_lons + 1, one_line_lons + 2]

    scanlines["telemetry"] = 2047
    scanlines["sensor_data"] = 99
    scanlines["add_on_zenith"] = 0
    scanlines["clock_drift_delta"] = 0
    scanlines["sensor_data"] = 2047

    pod_filename = tmp_path / "image.l1b"
    offset = 14800
    fill = np.zeros(offset - header3.itemsize, dtype=np.uint8)

    with open(pod_filename, "wb") as fd_:
        fd_.write(tbm_header.tobytes())
        fd_.write(header.tobytes())
        fd_.write(fill.tobytes())
        fd_.write(scanlines.tobytes())
    return pod_filename


@pytest.fixture()
def pod_tle(tmp_path):
    lines = ("1 23455U 94089A   00322.04713399  .00000318  00000-0  19705-3 0  5298\n"
             "2 23455  99.1591 303.5706 0010037  25.7760 334.3905 14.12496755303183\n"
             "1 23455U 94089A   00322.96799836  .00000229  00000-0  14918-3 0  5303\n"
             "2 23455  99.1590 304.5117 0009979  23.1101 337.0518 14.12496633303313\n")
    tle_filename = tmp_path / "noaa14.tle"
    with tle_filename.open("w") as fd:
        fd.write(lines)
    return tle_filename


def test_podlac_eosip(pod_file_with_tbm_header):
    """Test reading a real podlac file."""
    reader = LACPODReader(interpolate_coords=False)
    reader.read(pod_file_with_tbm_header)
    assert reader.head.itemsize == header3.itemsize
    # this is broken in eosip pod data, tbm data set name has start and end times reversed.
    # assert reader.head["data_set_name"] == reader.tbm_head["data_set_name"]
    # todo: test that duplicate lines are removed and recomputed


def test_read_to_dataset_is_a_dataset_including_channels_and_telemetry(pod_file_with_tbm_header, pod_tle):
    """Test creating an xr.Dataset from a gac file."""
    import xarray as xr
    reader = LACPODReader(interpolate_coords=False, tle_dir=pod_tle.parent, tle_name=pod_tle.name)
    dataset = reader.read_as_dataset(pod_file_with_tbm_header)
    assert isinstance(dataset, xr.Dataset)
    assert dataset["channels"].shape == (3, 2048, 5)
    assert dataset.attrs["processing_block_id"] == b"3031919"
    assert "times" in dataset.coords
    assert "scan_line_index" in dataset.coords
    assert "channel_name" in dataset.coords
    assert dataset["prt_counts"].shape == (3,)
    assert dataset["ict_counts"].shape == (3, 3)
    assert dataset["space_counts"].shape == (3, 3)


def test_read_to_dataset_without_interpolation(pod_file_with_tbm_header, pod_tle):
    """Test creating an xr.Dataset from a gac file."""
    reader = LACPODReader(interpolate_coords=False, tle_dir=pod_tle.parent, tle_name=pod_tle.name)
    dataset = reader.read_as_dataset(pod_file_with_tbm_header)

    assert dataset["longitude"].shape == (3, 51)
    assert dataset["latitude"].shape == (3, 51)


def test_read_to_dataset_with_interpolation(pod_file_with_tbm_header, pod_tle):
    """Test creating an xr.Dataset from a gac file."""
    reader = LACPODReader(interpolate_coords=True, tle_dir=pod_tle.parent, tle_name=pod_tle.name)
    dataset = reader.read_as_dataset(pod_file_with_tbm_header)

    assert dataset.coords["longitude"].shape == (3, 2048)
    assert dataset.coords["latitude"].shape == (3, 2048)


def test_passing_calibration_coeffs_to_reader_init_is_deprecated():
    """Test that passing calibration-specific info to the reader's init is not allowed anymore."""
    with pytest.warns(PendingDeprecationWarning):
        FakeGACReader(calibration_method="noaa", custom_calibration=dict(a="1"))
    with pytest.warns(PendingDeprecationWarning):
        FakeGACReader(calibration_method="noaa", calibration_file="somefile")


def test_passing_calibration_to_reader():
    """Test passing calibration info to `get_calibrated_channels`."""
    method = "InvalidMethod"
    with pytest.raises(ValueError, match=method):
        reader = FakeGACReader(calibration_method=method)

    reader = FakeGACReader(calibration_method="noaa")

    res = reader.get_calibrated_channels()

    np.testing.assert_allclose(res[:, 1, 0], 8.84714652)
    np.testing.assert_allclose(res[:, 2, 1], 10.23511303)
    assert reader.meta_data["calib_coeffs_version"] == "PATMOS-x, v2023"


def test_computing_lonlats(pod_file_with_tbm_header, pod_tle):
    """Test computing lons and lats from TLE data."""
    reader = LACPODReader(tle_dir=pod_tle.parent, tle_name=pod_tle.name, compute_lonlats_from_tles=True)
    dataset = reader.read_as_dataset(pod_file_with_tbm_header)
    lons = dataset["longitude"].values
    assert lons[0, 0] == pytest.approx(-4.756486)
    assert lons[0, -1] == pytest.approx(79.505688)


def test_recomputing_lonlats_with_time_offset(pod_file_with_tbm_header, pod_tle):
    """Test computing lons and lats from TLE data with time offset."""
    reader = LACPODReader(tle_dir=pod_tle.parent, tle_name=pod_tle.name, compute_lonlats_from_tles=True)
    reader.read(pod_file_with_tbm_header)
    lons, lats = reader._compute_lonlats(time_offset=np.timedelta64(500, "ms"))
    assert lons[0, 0] == pytest.approx(-4.714710005688344)
    assert lons[0, -1] == pytest.approx(79.44587709203307)


def test_georeferencing_with_first_guess(pod_file_with_tbm_header, pod_tle, monkeypatch):
    """Test getting a first guess for georeferencing from the file lon/lats."""
    expected_time_offset = 2  # seconds
    def mock_cal(counts, *args, **kwargs):
        return counts * 1.0
    import pygac.calibration.noaa
    monkeypatch.setattr(pygac.calibration.noaa, "calibrate_thermal", mock_cal)

    def mock_disp(*args):
        return 0, (0, 0, 0), ([10000], [1000])
    from georeferencer import georeferencer
    monkeypatch.setattr(georeferencer, "get_swath_displacement", mock_disp)
    reader = LACPODReader(tle_dir=pod_tle.parent, tle_name=pod_tle.name, compute_lonlats_from_tles=True,
                          reference_image="some_world_image.tif", adjust_clock_drift=False)
    def file_lonlats(self, *args, **kwargs):
        lons, lats = self._compute_lonlats(time_offset=np.timedelta64(int(expected_time_offset * 1e9), "ns"))
        return lons[:, self.lonlat_sample_points], lats[:, self.lonlat_sample_points]
    reader._get_lonlat_from_file = file_lonlats.__get__(reader)
    reader.read(pod_file_with_tbm_header)
    start_time = reader.get_times()[0]
    _ = reader.get_calibrated_dataset()
    new_start_time = reader.get_times()[0]
    tdiff = start_time - new_start_time + np.timedelta64(expected_time_offset, "s")
    assert tdiff < np.timedelta64(2, "ms")


def test_georeferencing_fails(pod_file_with_tbm_header, pod_tle, monkeypatch):
    """Test georeferencing."""

    def mock_cal(counts, *args, **kwargs):
        return counts * 1.0
    import pygac.calibration.noaa
    monkeypatch.setattr(pygac.calibration.noaa, "calibrate_thermal", mock_cal)

    def mock_disp(*args):
        return 0, (0, 0, 0), ([10000], [10000])
    from georeferencer import georeferencer
    monkeypatch.setattr(georeferencer, "get_swath_displacement", mock_disp)
    reader = LACPODReader(tle_dir=pod_tle.parent, tle_name=pod_tle.name, compute_lonlats_from_tles=True,
                          reference_image="some_world_image.tif")
    reader.read(pod_file_with_tbm_header)
    with pytest.raises(RuntimeError):
        _ = reader.get_calibrated_dataset()


def test_georeferencing(pod_file_with_tbm_header, pod_tle, monkeypatch):
    """Test georeferencing."""

    def mock_cal(counts, *args, **kwargs):
        return counts * 1.0
    import pygac.calibration.noaa
    monkeypatch.setattr(pygac.calibration.noaa, "calibrate_thermal", mock_cal)

    def mock_disp(*args):
        return 0.5, (0, 0, 0), ([10000], [1000])
    from georeferencer import georeferencer
    monkeypatch.setattr(georeferencer, "get_swath_displacement", mock_disp)
    reader = LACPODReader(tle_dir=pod_tle.parent, tle_name=pod_tle.name, compute_lonlats_from_tles=True,
                          reference_image="some_world_image.tif")
    reader.read(pod_file_with_tbm_header)
    dataset = reader.get_calibrated_dataset()
    assert dataset.attrs["max_scan_angle"] == 55.37
    lons = dataset["longitude"].values
    assert lons[0, 0] == pytest.approx(-4.714710005688344)
    assert lons[0, -1] == pytest.approx(79.44587709203307)

def test_orthocorrection(pod_file_with_tbm_header, pod_tle, monkeypatch):
    """Test computing lons and lats from TLE data."""

    def mock_cal(counts, *args, **kwargs):
        return counts * 1.0
    import pygac.calibration.noaa
    monkeypatch.setattr(pygac.calibration.noaa, "calibrate_thermal", mock_cal)

    def mock_disp(*args):
        return 0.5, (0, 0, 0), ([10000], [1000])
    from georeferencer import georeferencer
    monkeypatch.setattr(georeferencer, "get_swath_displacement", mock_disp)
    reader = LACPODReader(tle_dir=pod_tle.parent, tle_name=pod_tle.name, compute_lonlats_from_tles=True,
                          reference_image="some_world_image.tif", dem="dem_file.tif")
    reader.read(pod_file_with_tbm_header)


    def mock_orthocorrection(calibrated_ds, *args):
        calibrated_ds["tc_lons"] = calibrated_ds["longitude"] + 0.0001
        calibrated_ds["tc_lats"] = calibrated_ds["latitude"] + 0.0001
        return calibrated_ds

    monkeypatch.setattr(georeferencer, "orthocorrection", mock_orthocorrection)
    dataset = reader.get_calibrated_dataset()

    original_lons = dataset["longitude"].copy()
    original_lats = dataset["latitude"].copy()

    dataset = georeferencer.orthocorrection(dataset)
    assert np.array_equal(dataset["longitude"], original_lons)
    assert np.array_equal(dataset["latitude"], original_lats)

    assert "tc_lons" in dataset
    assert "tc_lats" in dataset

    assert dataset["tc_lons"].shape == dataset["longitude"].shape
    assert dataset["tc_lats"].shape == dataset["latitude"].shape

    assert not np.array_equal(dataset["tc_lons"], dataset["longitude"])
    assert not np.array_equal(dataset["tc_lats"], dataset["latitude"])

def test_read_tle_file(pod_tle, tmp_path):
    tle_content = """1 33591U 09005A   24076.18425395  .00000218  00000+0  14176-3 0  9993
2 33591  99.0596 130.9575 0013809 190.5723 169.5160 14.12946284778493
1 33591U 09005A   24077.17564174  .00000205  00000+0  13478-3 0  9992
2 33591  99.0594 131.9606 0013864 187.8089 172.2868 14.12946798778634

"""
    tle_file = tmp_path / "test.tle"
    tle_file.write_text(tle_content)

    reader = LACPODReader(
        tle_dir=pod_tle.parent,
        tle_name=pod_tle.name,
        compute_lonlats_from_tles=True,
    )
    result = reader.read_tle_file(tle_file)

    expected = [
        "1 33591U 09005A   24076.18425395  .00000218  00000+0  14176-3 0  9993\n",
        "2 33591  99.0596 130.9575 0013809 190.5723 169.5160 14.12946284778493\n",
        "1 33591U 09005A   24077.17564174  .00000205  00000+0  13478-3 0  9992\n",
        "2 33591  99.0594 131.9606 0013864 187.8089 172.2868 14.12946798778634\n",
    ]
    assert result == expected



"""Reader for GAC KLM data."""

from __future__ import print_function

import logging

import numpy as np

from pygac.gac_reader import GACReader
from pygac.klm_reader import KLMReader, main_klm

LOG = logging.getLogger(__name__)

# video data object

scanline = np.dtype([("scan_line_number", ">u2"),
                     ("scan_line_year", ">u2"),
                     ("scan_line_day_of_year", ">u2"),
                     ("satellite_clock_drift_delta", ">i2"),
                     ("scan_line_utc_time_of_day", ">u4"),
                     ("scan_line_bit_field", ">u2"),
                     ("zero_fill0", ">i2", (5, )),
                     # QUALITY INDICATORS
                     ("quality_indicator_bit_field", ">u4"),
                     ("scan_line_quality_flags", [("reserved", ">u1"),
                                                  ("time_problem_code", ">u1"),
                                                  ("calibration_problem_code", ">u1"),
                                                  ("earth_location_problem_code", ">u1")]),
                     ("calibration_quality_flags", ">u2", (3, )),
                     ("count_of_bit_errors_in_frame_sync", ">u2"),
                     ("zero_fill1", ">i4", (2, )),
                     # CALIBRATION COEFFICIENTS
                     ("visible_operational_cal_ch_1_slope_1", ">i4"),
                     ("visible_operational_cal_ch_1_intercept_1", ">i4"),
                     ("visible_operational_cal_ch_1_slope_2", ">i4"),
                     ("visible_operational_cal_ch_1_intercept_2", ">i4"),
                     ("visible_operational_cal_ch_1_intersection", ">i4"),
                     ("visible_test_cal_ch_1_slope_1", ">i4"),
                     ("visible_test_cal_ch_1_intercept_1", ">i4"),
                     ("visible_test_cal_ch_1_slope_2", ">i4"),
                     ("visible_test_cal_ch_1_intercept_2", ">i4"),
                     ("visible_test_cal_ch_1_intersection", ">i4"),
                     ("visible_prelaunch_cal_ch_1_slope_1", ">i4"),
                     ("visible_prelaunch_cal_ch_1_intercept_1", ">i4"),
                     ("visible_prelaunch_cal_ch_1_slope_2", ">i4"),
                     ("visible_prelaunch_cal_ch_1_intercept_2", ">i4"),
                     ("visible_prelaunch_cal_ch_1_intersection", ">i4"),
                     ("visible_operational_cal_ch_2_slope_1", ">i4"),
                     ("visible_operational_cal_ch_2_intercept_1", ">i4"),
                     ("visible_operational_cal_ch_2_slope_2", ">i4"),
                     ("visible_operational_cal_ch_2_intercept_2", ">i4"),
                     ("visible_operational_cal_ch_2_intersection", ">i4"),
                     ("visible_test_cal_ch_2_slope_1", ">i4"),
                     ("visible_test_cal_ch_2_intercept_1", ">i4"),
                     ("visible_test_cal_ch_2_slope_2", ">i4"),
                     ("visible_test_cal_ch_2_intercept_2", ">i4"),
                     ("visible_test_cal_ch_2_intersection", ">i4"),
                     ("visible_prelaunch_cal_ch_2_slope_1", ">i4"),
                     ("visible_prelaunch_cal_ch_2_intercept_1", ">i4"),
                     ("visible_prelaunch_cal_ch_2_slope_2", ">i4"),
                     ("visible_prelaunch_cal_ch_2_intercept_2", ">i4"),
                     ("visible_prelaunch_cal_ch_2_intersection", ">i4"),
                     ("visible_operational_cal_ch_3a_slope_1", ">i4"),
                     ("visible_operational_cal_ch_3a_intercept_1", ">i4"),
                     ("visible_operational_cal_ch_3a_slope_2", ">i4"),
                     ("visible_operational_cal_ch_3a_intercept_2", ">i4"),
                     ("visible_operational_cal_ch_3a_intersection", ">i4"),
                     ("visible_test_cal_ch_3a_slope_1", ">i4"),
                     ("visible_test_cal_ch_3a_intercept_1", ">i4"),
                     ("visible_test_cal_ch_3a_slope_2", ">i4"),
                     ("visible_test_cal_ch_3a_intercept_2", ">i4"),
                     ("visible_test_cal_ch_3a_intersection", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_slope_1", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_intercept_1", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_slope_2", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_intercept_2", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_intersection", ">i4"),
                     ("ir_operational_cal_ch_3b_coefficient_1", ">i4"),
                     ("ir_operational_cal_ch_3b_coefficient_2", ">i4"),
                     ("ir_operational_cal_ch_3b_coefficient_3", ">i4"),
                     ("ir_test_cal_ch_3b_coefficient_1", ">i4"),
                     ("ir_test_cal_ch_3b_coefficient_2", ">i4"),
                     ("ir_test_cal_ch_3b_coefficient_3", ">i4"),
                     ("ir_operational_cal_ch_4_coefficient_1", ">i4"),
                     ("ir_operational_cal_ch_4_coefficient_2", ">i4"),
                     ("ir_operational_cal_ch_4_coefficient_3", ">i4"),
                     ("ir_test_cal_ch_4_coefficient_1", ">i4"),
                     ("ir_test_cal_ch_4_coefficient_2", ">i4"),
                     ("ir_test_cal_ch_4_coefficient_3", ">i4"),
                     ("ir_operational_cal_ch_5_coefficient_1", ">i4"),
                     ("ir_operational_cal_ch_5_coefficient_2", ">i4"),
                     ("ir_operational_cal_ch_5_coefficient_3", ">i4"),
                     ("ir_test_cal_ch_5_coefficient_1", ">i4"),
                     ("ir_test_cal_ch_5_coefficient_2", ">i4"),
                     ("ir_test_cal_ch_5_coefficient_3", ">i4"),
                     # NAVIGATION
                     ("computed_yaw_steering", ">i2", (3,)),  # only in version 5
                     ("total_applied_attitude_correction",
                      ">i2", (3,)),  # only in version 5
                     ("navigation_status_bit_field", ">u4"),
                     ("time_associated_with_tip_euler_angles", ">u4"),
                     ("tip_euler_angles", ">i2", (3, )),
                     ("spacecraft_altitude_above_reference_ellipsoid", ">u2"),
                     ("angular_relationships", ">i2", (153, )),
                     ("zero_fill3", ">i2", (3, )),
                     ("earth_location", [("lats", ">i4"),
                                         ("lons", ">i4")], (51,)),
                     ("zero_fill4", ">i4", (2, )),
                     # HRPT MINOR FRAME TELEMETRY
                     ("frame_sync", ">u2", (6, )),
                     ("id", ">u2", (2, )),
                     ("time_code", ">u2", (4, )),
                     ("telemetry", [("ramp_calibration", ">u2", (5, )),
                                    ("PRT", ">u2", (3, )),
                                    ("ch3_patch_temp", ">u2"),
                                    ("spare", ">u2"), ]),
                     ("back_scan", ">u2", (30, )),
                     ("space_data", ">u2", (50, )),
                     ("sync_delta", ">u2"),
                     ("zero_fill5", ">i2"),
                     # AVHRR SENSOR DATA
                     ("sensor_data", ">u4", (682, )),
                     ("zero_fill6", ">i4", (2, )),
                     # DIGITAL B TELEMETRY
                     ("invalid_word_bit_flags1", ">u2"),
                     ("avhrr_digital_b_data", ">u2"),
                     ("zero_fill7", ">i4", (3, )),
                     # ANALOG HOUSEKEEPING DATA (TIP)
                     ("invalid_word_bit_flags2", ">u4"),
                     ("patch_temperature_range", ">u1"),
                     ("patch_temperature_extended", ">u1"),
                     ("patch_power", ">u1"),
                     ("radiator_temperature", ">u1"),
                     ("blackbody_temperature_1", ">u1"),
                     ("blackbody_temperature_2", ">u1"),
                     ("blackbody_temperature_3", ">u1"),
                     ("blackbody_temperature_4", ">u1"),
                     ("electronics_current", ">u1"),
                     ("motor_current", ">u1"),
                     ("earth_shield_position", ">u1"),
                     ("electronics_temperature", ">u1"),
                     ("cooler_housing_temperature", ">u1"),
                     ("baseplate_temperature", ">u1"),
                     ("motor_housing_temperature", ">u1"),
                     ("a/d_converter_temperature", ">u1"),
                     ("detector_#4_bias_voltage", ">u1"),
                     ("detector_#5_bias_voltage", ">u1"),
                     ("blackbody_temperature_channel3b", ">u1"),
                     ("blackbody_temperature_channel4", ">u1"),
                     ("blackbody_temperature_channel5", ">u1"),
                     ("reference_voltage", ">u1"),
                     ("zero_fill8", ">i2", (3, )),
                     # CLOUDS FROM AVHRR (CLAVR)
                     ("reserved0", ">u4"),
                     ("reserved1", ">u4"),
                     ("reserved2", ">u2", (52, )),
                     # FILLER
                     ("zero_fill9", ">i4", (112, ))])


class GACKLMReader(GACReader, KLMReader):
    """The GAC KLM reader class.

    The offset attribute tells where in the file the scanline data starts.
    """
    lonlat_sample_points = np.arange(4.5, 405, 8)

    def __init__(self, *args, **kwargs):
        """Init the GAC KLM reader."""
        GACReader.__init__(self, *args, **kwargs)
        self.scanline_type = scanline
        self.offset = 4608


def main(filename, start_line, end_line):
    """Generate a l1c file."""
    return main_klm(GACKLMReader, filename, start_line, end_line)


if __name__ == "__main__":
    import sys
    main(sys.argv[1], sys.argv[2], sys.argv[3])



"""Test pygac.utils module
"""

import gzip
import io
import os
import unittest
from unittest import mock

import numpy as np

from pygac.utils import calculate_sun_earth_distance_correction, file_opener


class TestUtils(unittest.TestCase):
    """Test pygac.utils functions"""

    longMessage = True

    @mock.patch("pygac.utils.open", mock.MagicMock(return_value=io.BytesIO(b"file content")))
    def test_file_opener_1(self):
        """Test if a file is redirected correctly through file_opener."""
        with file_opener("path/to/file") as f:
            content = f.read()
        self.assertEqual(content, b"file content")

    def test_file_opener_2(self):
        """Test file_opener with file objects and compression"""
        # prepare test
        normal_message = b"normal message"
        gzip_message_decoded = b"gzip message"
        with io.BytesIO() as f:
            with gzip.open(f, mode="wb") as g:
                g.write(gzip_message_decoded)
            f.seek(0)
            gzip_message_encoded = f.read()
        # on normal file (check also if it remains open)
        with io.BytesIO(normal_message) as f:
            with file_opener(f) as g:
                message = g.read()
            self.assertFalse(f.closed)
        self.assertEqual(message, normal_message)
        # on gzip file
        with io.BytesIO(gzip_message_encoded) as f:
            with file_opener(f) as g:
                message = g.read()
        self.assertEqual(message, gzip_message_decoded)

    @mock.patch("pygac.utils.open", mock.MagicMock(side_effect=FileNotFoundError))
    def test_file_opener_3(self):
        """Test file_opener with PathLike object"""
        # prepare test
        class RawBytes(os.PathLike):
            def __init__(self, filename, raw_bytes):
                self.filename = str(filename)
                self.raw_bytes = raw_bytes

            def __fspath__(self):
                return self.filename

            def open(self):
                return io.BytesIO(self.raw_bytes)

        filename = "/path/to/file"
        file_bytes = b"TestTestTest"
        test_pathlike = RawBytes(filename, file_bytes)
        with file_opener(test_pathlike) as f:
            content = f.read()
        self.assertEqual(content, file_bytes)

        # test with lazy loading open method (open only in context)
        class RawBytesLazy(RawBytes):
            def open(self):
                self.lazy_opener_mock = mock.MagicMock()
                self.lazy_opener_mock.__enter__.return_value = io.BytesIO(self.raw_bytes)
                return self.lazy_opener_mock

        test_pathlike = RawBytesLazy(filename, file_bytes)
        with file_opener(test_pathlike) as f:
            content = f.read()
        self.assertEqual(content, file_bytes)
        test_pathlike.lazy_opener_mock.__exit__.assert_called_once_with(None, None, None)

    def test_calculate_sun_earth_distance_correction(self):
        """Test function for the sun distance corretction."""
        corr = calculate_sun_earth_distance_correction(3)
        np.testing.assert_almost_equal(corr, 0.96660494, decimal=7)



"""Read KLM data.

Reads L1b GAC/LAC data from KLM series of satellites (NOAA-15 and later).
Format specification can be found in section 8 of the `KLM user guide`_.

.. _KLM user guide:
    https://www.ncei.noaa.gov/pub/data/satellite/publications/podguides/N-15%20thru%20N-19/

"""

import datetime
import logging
from enum import IntFlag

import numpy as np

from pygac.correct_tsm_issue import TSM_AFFECTED_INTERVALS_KLM, get_tsm_idx
from pygac.reader import Reader, ReaderError
from pygac.utils import file_opener

LOG = logging.getLogger(__name__)


class KLM_QualityIndicator(IntFlag):
    """Quality Indicators.

    Source: KLM guide

    - Table 8.3.1.3.3.1-1. Format of packed LAC/HRPT Data Sets (Version 2,
      pre-April 28, 2005).
    - Table 8.3.1.3.3.2-1. Format of LAC/HRPT Data Record for NOAA-N
      (Version 5, post-November 14, 2006, all spacecraft).
    - Table 8.3.1.4.3.1-1. Format of packed GAC Data Record for NOAA KLM
      (Version 2, pre-April 28, 2005).
    - Table 8.3.1.4.3.2-1. Format of GAC Data Record for NOAA-N (Version 4,
      post-January 25, 2006, all spacecraft).

    Notes:
      - Table 8.3.1.3.3.1-1. and Table 8.3.1.4.3.1-1. define bit: 21 as
        "frame sync word not valid"
      - Table 8.3.1.3.3.2-1. and Table 8.3.1.4.3.2-1. define bit: 21 as
        "flywheeling detected during this frame"

    """

    FATAL_FLAG = 2**31  # Data should not be used for product generation
    TIME_ERROR = 2**30  # Time sequence error detected within this scan
    DATA_GAP = 2**29  # Data gap precedes this scan
    CALIBRATION = 2**28  # Insufficient data for calibration
    NO_EARTH_LOCATION = 2**27  # Earth location data not available
    CLOCK_UPDATE = 2**26  # First good time following a clock update (nominally 0)
    INSTRUMENT_CHANGE = 2**25  # Instrument status changed with this scan
    BIT_SYNC_STATUS = 2**24  # Sync lock dropped during this frame
    SYNC_ERROR = 2**23  # Frame sync word error greater than zero
    FRAME_SYNC_LOCK = 2**22  # Frame sync previously dropped lock
    SYNC_INVALID = 2**21  # Frame sync word not valid
    FLYWHEELING = 2**21  # Flywheeling detected during this frame
    BIT_SLIPPAGE = 2**20  # Bit slippage detected during this frame
    # Note: Bit 19 - 9 are not defined for KLMs
    TIP_PARITY = 2**8  # TIP parity error detected
    # Reflected Sunlight (RS) detected (solar blackbody contamination)
    CH_3B_RS = 2**7
    CH_3B_RS_ANOMALY = 2**6
    CH_3_CONTAMINATION = CH_3B_RS | CH_3B_RS_ANOMALY  # POD compatible alias
    CH_4_RS = 2**5
    CH_4_RS_ANOMALY = 2**4
    CH_4_CONTAMINATION = CH_4_RS | CH_4_RS_ANOMALY  # POD compatible alias
    CH_5_RS = 2**3
    CH_5_RS_ANOMALY = 2**2
    CH_5_CONTAMINATION = CH_5_RS | CH_5_RS_ANOMALY  # POD compatible alias
    DATA_JITTER = 2**1  # Resync occurred on this frame
    PSEUDO_NOISE = 2**0  # Pseudo noise occurred on this frame


# header object

header = np.dtype([("data_set_creation_site_id", "S3"),
                   ("ascii_blank_=_x20", "S1"),
                   ("noaa_level_1b_format_version_number", ">u2"),
                   ("noaa_level_1b_format_version_year", ">u2"),
                   ("noaa_level_1b_format_version_day_of_year", ">u2"),
                   ("reserved_for_logical_record_length", ">u2"),
                   ("reserved_for_block_size", ">u2"),
                   ("count_of_header_records", ">u2"),
                   ("zero_fill0", ">i2", (3, )),
                   ("data_set_name", "S42"),
                   ("processing_block_identification", "S8"),
                   ("noaa_spacecraft_identification_code", ">u2"),
                   ("instrument_id", ">u2"),
                   ("data_type_code", ">u2"),
                   ("tip_source_code", ">u2"),
                   ("start_of_data_set_day_count_starting_from_0_at_00h,_1_jan_1950",
                    ">u4"),
                   ("start_of_data_set_year", ">u2"),
                   ("start_of_data_set_day_of_year", ">u2"),
                   ("start_of_data_set_utc_time_of_day", ">u4"),
                   ("end_of_data_set_day_count_starting_from_0_at_00h,_1_jan_1950",
                    ">u4"),
                   ("end_of_data_set_year", ">u2"),
                   ("end_of_data_set_day_of_year", ">u2"),
                   ("end_of_data_set_utc_time_of_day", ">u4"),
                   ("year_of_last_cpids_update", ">u2"),
                   ("day_of_year_of_last_cpids_update", ">u2"),
                   ("zero_fill1", ">i2", (4, )),
                   # DATA SET QUALITY INDICATORS
                   ("instrument_status", ">u4"),
                   ("zero_fill2", ">i2"),
                   ("record_number_of_status_change", ">u2"),
                   ("second_instrument_status", ">u4"),
                   ("count_of_data_records", ">u2"),
                   ("count_of_calibrated,_earth_located_scan_lines", ">u2"),
                   ("count_of_missing_scan_lines", ">u2"),
                   ("count_of_data_gaps", ">u2"),
                   # only in version 5
                   ("count_of_data_frames_without_frame_sync_word", ">u2"),
                   ("count_of_pacs_detected_tip_parity_errors", ">u2"),
                   # only in version 5
                   ("sum_of_all_auxiliary_sync_errors_detected", ">u2"),
                   ("time_sequence_error", ">u2"),
                   ("time_sequence_error_code", ">u2"),
                   ("socc_clock_update_indicator", ">u2"),
                   ("earth_location_error_indicator", ">u2"),
                   ("earth_location_error_code", ">u2"),
                   ("pacs_status_bit_field", ">u2"),
                   ("data_source", ">u2"),
                   ("zero_fill3", ">i4"),
                   ("reserved_for_the_ingester", "S8"),
                   ("reserved_for_decommutation", "S8"),
                   ("zero_fill4", ">i2", (5, )),
                   # CALIBRATION
                   ("ramp_auto_calibration_indicators_bit_field", ">u2"),
                   ("year_of_most_recent_solar_channel_calibration", ">u2"),
                   ("day_of_year_of_most_recent_solar_channel_calibration",
                    ">u2"),
                   ("primary_calibration_algorithm_id", ">u2"),
                   ("primary_calibration_algorithm_selected_options", ">u2"),
                   ("secondary_calibration_algorithm_id", ">u2"),
                   ("secondary_calibration_algorithm_selected_options", ">u2"),
                   ("ir_target_temperature_1_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_6", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_6", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_6", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_6", ">i2"),
                   ("zero_fill5", ">i4", (2, )),
                   # RADIANCE CONVERSION
                   ("ch_1_solar_filtered_irradiance_in_wavelength", ">i4"),
                   ("ch_1_equivalent_filter_width_in_wavelength", ">i4"),
                   ("ch_2_solar_filtered_irradiance_in_wavelength", ">i4"),
                   ("ch_2_equivalent_filter_width_in_wavelength", ">i4"),
                   ("ch_3a_solar_filtered_irradiance_in_wavelength", ">i4"),
                   ("ch_3a_equivalent_filter_width_in_wavelength", ">i4"),
                   ("ch_3b_central_wavenumber", ">i4"),
                   ("ch_3b_constant_1", ">i4"),
                   ("ch_3b_constant_2", ">i4"),
                   ("ch_4_central_wavenumber", ">i4"),
                   ("ch_4_constant_1", ">i4"),
                   ("ch_4_constant_2", ">i4"),
                   ("ch_5_central_wavenumber", ">i4"),
                   ("ch_5_constant_1", ">i4"),
                   ("ch_5_constant_2", ">i4"),
                   ("zero_fill6", ">i4", (3, )),
                   # NAVIGATION
                   ("reference_ellipsoid_model_id", "S8"),
                   ("nadir_earth_location_tolerance", ">u2"),
                   ("earth_location_bit_field", ">u2"),
                   ("zero_fill7", ">i2"),
                   ("constant_roll_attitude_error", ">i2"),
                   ("constant_pitch_attitude_error", ">i2"),
                   ("constant_yaw_attitude_error", ">i2"),
                   ("epoch_year_for_orbit_vector", ">u2"),
                   ("day_of_epoch_year_for_orbit_vector", ">u2"),
                   ("epoch_utc_time_of_day_for_orbit_vector", ">u4"),
                   ("semi-major_axis", ">i4"),
                   ("eccentricity", ">i4"),
                   ("inclination", ">i4"),
                   ("argument_of_perigee", ">i4"),
                   ("right_ascension_of_the_ascending_node", ">i4"),
                   ("mean_anomaly", ">i4"),
                   ("position_vector_x_component", ">i4"),
                   ("position_vector_y_component", ">i4"),
                   ("position_vector_z_component", ">i4"),
                   ("velocity_vector_x-dot_component", ">i4"),
                   ("velocity_vector_y-dot_component", ">i4"),
                   ("velocity_vector_z-dot_component", ">i4"),
                   ("earth_sun_distance_ratio", ">u4"),
                   ("zero_fill8", ">i4", (4, ))])

# ANALOG TELEMETRY CONVERSION
analog_telemetry_v2 = np.dtype([("patch_temperature_conversion_coefficient_1", ">i2"),
                                ("patch_temperature_conversion_coefficient_2", ">i2"),
                                ("patch_temperature_conversion_coefficient_3", ">i2"),
                                ("patch_temperature_conversion_coefficient_4", ">i2"),
                                ("patch_temperature_conversion_coefficient_5", ">i2"),
                                ("reserved0", ">i2"),
                                ("patch_temperature_extended_conversion_coefficient_1",
                                 ">i2"),
                                ("patch_temperature_extended_conversion_coefficient_2",
                                 ">i2"),
                                ("patch_temperature_extended_conversion_coefficient_3",
                                 ">i2"),
                                ("patch_temperature_extended_conversion_coefficient_4",
                                 ">i2"),
                                ("patch_temperature_extended_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved1", ">i2"),
                                ("patch_power_conversion_coefficient_1", ">i2"),
                                ("patch_power_conversion_coefficient_2", ">i2"),
                                ("patch_power_conversion_coefficient_3", ">i2"),
                                ("patch_power_conversion_coefficient_4", ">i2"),
                                ("patch_power_conversion_coefficient_5", ">i2"),
                                ("reserved2", ">i2"),
                                ("radiator_temperature_conversion_coefficient_1", ">i2"),
                                ("radiator_temperature_conversion_coefficient_2", ">i2"),
                                ("radiator_temperature_conversion_coefficient_3", ">i2"),
                                ("radiator_temperature_conversion_coefficient_4", ">i2"),
                                ("radiato   r_temperature_conversion_coefficient_5", ">i2"),
                                ("reserved3", ">i2"),
                                ("blackbody_temperature_1_conversion_coefficient_1", ">i2"),
                                ("blackbody_temperature_1_conversion_coefficient_2", ">i2"),
                                ("blackbody_temperature_1_conversion_coefficient_3", ">i2"),
                                ("blackbody_temperature_1_conversion_coefficient_4", ">i2"),
                                ("blackbody_temperature_1_conversion_coefficient_5", ">i2"),
                                ("reserved4", ">i2"),
                                ("blackbody_temperature_2_conversion_coefficient_1", ">i2"),
                                ("blackbody_temperature_2_conversion_coefficient_2", ">i2"),
                                ("blackbody_temperature_2_conversion_coefficient_3", ">i2"),
                                ("blackbody_temperature_2_conversion_coefficient_4", ">i2"),
                                ("blackbody_temperature_2_conversion_coefficient_5", ">i2"),
                                ("reserved5", ">i2"),
                                ("blackbody_temperature_3_conversion_coefficient_1", ">i2"),
                                ("blackbody_temperature_3_conversion_coefficient_2", ">i2"),
                                ("blackbody_temperature_3_conversion_coefficient_3", ">i2"),
                                ("blackbody_temperature_3_conversion_coefficient_4", ">i2"),
                                ("blackbody_temperature_3_conversion_coefficient_5", ">i2"),
                                ("reserved6", ">i2"),
                                ("blackbody_temperature_4_conversion_coefficient_1", ">i2"),
                                ("blackbody_temperature_4_conversion_coefficient_2", ">i2"),
                                ("blackbody_temperature_4_conversion_coefficient_3", ">i2"),
                                ("blackbody_temperature_4_conversion_coefficient_4", ">i2"),
                                ("blackbody_temperature_4_conversion_coefficient_5", ">i2"),
                                ("reserved7", ">i2"),
                                ("electronics_current_conversion_coefficient_1", ">i2"),
                                ("electronics_current_conversion_coefficient_2", ">i2"),
                                ("electronics_current_conversion_coefficient_3", ">i2"),
                                ("electronics_current_conversion_coefficient_4", ">i2"),
                                ("electronics_current_conversion_coefficient_5", ">i2"),
                                ("reserved8", ">i2"),
                                ("motor_current_conversion_coefficient_1", ">i2"),
                                ("motor_current_conversion_coefficient_2", ">i2"),
                                ("motor_current_conversion_coefficient_3", ">i2"),
                                ("motor_current_conversion_coefficient_4", ">i2"),
                                ("motor_current_conversion_coefficient_5", ">i2"),
                                ("reserved9", ">i2"),
                                ("earth_shield_position_conversion_coefficient_1", ">i2"),
                                ("earth_shield_position_conversion_coefficient_2", ">i2"),
                                ("earth_shield_position_conversion_coefficient_3", ">i2"),
                                ("earth_shield_position_conversion_coefficient_4", ">i2"),
                                ("earth_shield_position_conversion_coefficient_5", ">i2"),
                                ("reserved10", ">i2"),
                                ("electronics_temperature_conversion_coefficient_1", ">i2"),
                                ("electronics_temperature_conversion_coefficient_2", ">i2"),
                                ("electronics_temperature_conversion_coefficient_3", ">i2"),
                                ("electronics_temperature_conversion_coefficient_4", ">i2"),
                                ("electronics_temperature_conversion_coefficient_5", ">i2"),
                                ("reserved11", ">i2"),
                                ("cooler_housing_temperature_conversion_coefficient_1",
                                 ">i2"),
                                ("cooler_housing_temperature_conversion_coefficient_2",
                                 ">i2"),
                                ("cooler_housing_temperature_conversion_coefficient_3",
                                 ">i2"),
                                ("cooler_housing_temperature_conversion_coefficient_4",
                                 ">i2"),
                                ("cooler_housing_temperature_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved12", ">i2"),
                                ("baseplate_temperature_conversion_coefficient_1", ">i2"),
                                ("baseplate_temperature_conversion_coefficient_2", ">i2"),
                                ("baseplate_temperature_conversion_coefficient_3", ">i2"),
                                ("baseplate_temperature_conversion_coefficient_4", ">i2"),
                                ("baseplate_temperature_conversion_coefficient_5", ">i2"),
                                ("reserved13", ">i2"),
                                ("motor_housing_temperature_conversion_coefficient_1",
                                 ">i2"),
                                ("motor_housing_temperature_conversion_coefficient_2",
                                 ">i2"),
                                ("motor_housing_temperature_conversion_coefficient_3",
                                 ">i2"),
                                ("motor_housing_temperature_conversion_coefficient_4",
                                 ">i2"),
                                ("motor_housing_temperature_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved14", ">i2"),
                                ("a/d_converter_temperature_conversion_coefficient_1",
                                 ">i2"),
                                ("a/d_converter_temperature_conversion_coefficient_2",
                                 ">i2"),
                                ("a/d_converter_temperature_conversion_coefficient_3",
                                 ">i2"),
                                ("a/d_converter_temperature_conversion_coefficient_4",
                                 ">i2"),
                                ("a/d_converter_temperature_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved15", ">i2"),
                                ("detector_#4_bias_voltage_conversion_coefficient_1",
                                 ">i2"),
                                ("detector_#4_bias_voltage_conversion_coefficient_2",
                                 ">i2"),
                                ("detector_#4_bias_voltage_conversion_coefficient_3",
                                 ">i2"),
                                ("detector_#4_bias_voltage_conversion_coefficient_4",
                                 ">i2"),
                                ("detector_#4_bias_voltage_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved16", ">i2"),
                                ("detector_#5_bias_voltage_conversion_coefficient_1",
                                 ">i2"),
                                ("detector_#5_bias_voltage_conversion_coefficient_2",
                                 ">i2"),
                                ("detector_#5_bias_voltage_conversion_coefficient_3",
                                 ">i2"),
                                ("detector_#5_bias_voltage_conversion_coefficient_4",
                                 ">i2"),
                                ("detector_#5_bias_voltage_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved17", ">i2"),
                                ("channel_3b_blackbody_view_conversion_coefficient_1",
                                 ">i2"),
                                ("channel_3b_blackbody_view_conversion_coefficient_2",
                                 ">i2"),
                                ("channel_3b_blackbody_view_conversion_coefficient_3",
                                 ">i2"),
                                ("channel_3b_blackbody_view_conversion_coefficient_4",
                                 ">i2"),
                                ("channel_3b_blackbody_view_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved18", ">i2"),
                                ("channel_4_blackbody_view_conversion_coefficient_1",
                                 ">i2"),
                                ("channel_4_blackbody_view_conversion_coefficient_2",
                                 ">i2"),
                                ("channel_4_blackbody_view_conversion_coefficient_3",
                                 ">i2"),
                                ("channel_4_blackbody_view_conversion_coefficient_4",
                                 ">i2"),
                                ("channel_4_blackbody_view_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved19", ">i2"),
                                ("channel_5_blackbody_view_conversion_coefficient_1",
                                 ">i2"),
                                ("channel_5_blackbody_view_conversion_coefficient_2",
                                 ">i2"),
                                ("channel_5_blackbody_view_conversion_coefficient_3",
                                 ">i2"),
                                ("channel_5_blackbody_view_conversion_coefficient_4",
                                 ">i2"),
                                ("channel_5_blackbody_view_conversion_coefficient_5",
                                 ">i2"),
                                ("reserved20", ">i2"),
                                ("reference_voltage_conversion_coefficient_1", ">i2"),
                                ("reference_voltage_conversion_coefficient_2", ">i2"),
                                ("reference_voltage_conversion_coefficient_3", ">i2"),
                                ("reference_voltage_conversion_coefficient_4", ">i2"),
                                ("reference_voltage_conversion_coefficient_5", ">i2"),
                                ("reserved21", ">i2")])
# FILLER
#                  ("zero_fill9", ">i4", (4608 - 688, ))])


analog_telemetry_v5 = np.dtype([("patch_temperature_conversion_coefficient_1", ">i4"),
                                ("patch_temperature_conversion_coefficient_2", ">i4"),
                                ("patch_temperature_conversion_coefficient_3", ">i4"),
                                ("patch_temperature_conversion_coefficient_4", ">i4"),
                                ("patch_temperature_conversion_coefficient_5", ">i4"),
                                ("patch_temperature_conversion_coefficient_6", ">i4"),
                                ("patch_temperature_extended_conversion_coefficient_1",
                                 ">i4"),
                                ("patch_temperature_extended_conversion_coefficient_2",
                                 ">i4"),
                                ("patch_temperature_extended_conversion_coefficient_3",
                                 ">i4"),
                                ("patch_temperature_extended_conversion_coefficient_4",
                                 ">i4"),
                                ("patch_temperature_extended_conversion_coefficient_5",
                                 ">i4"),
                                ("patch_temperature_extended_conversion_coefficient_6", ">i4"),
                                ("patch_power_conversion_coefficient_1", ">i4"),
                                ("patch_power_conversion_coefficient_2", ">i4"),
                                ("patch_power_conversion_coefficient_3", ">i4"),
                                ("patch_power_conversion_coefficient_4", ">i4"),
                                ("patch_power_conversion_coefficient_5", ">i4"),
                                ("patch_power_conversion_coefficient_6", ">i4"),
                                ("radiator_temperature_conversion_coefficient_1", ">i4"),
                                ("radiator_temperature_conversion_coefficient_2", ">i4"),
                                ("radiator_temperature_conversion_coefficient_3", ">i4"),
                                ("radiator_temperature_conversion_coefficient_4", ">i4"),
                                ("radiator_temperature_conversion_coefficient_5", ">i4"),
                                ("radiator_temperature_conversion_coefficient_6", ">i4"),
                                ("blackbody_temperature_1_conversion_coefficient_1", ">i4"),
                                ("blackbody_temperature_1_conversion_coefficient_2", ">i4"),
                                ("blackbody_temperature_1_conversion_coefficient_3", ">i4"),
                                ("blackbody_temperature_1_conversion_coefficient_4", ">i4"),
                                ("blackbody_temperature_1_conversion_coefficient_5", ">i4"),
                                ("blackbody_temperature_1_conversion_coefficient_6", ">i4"),
                                ("blackbody_temperature_2_conversion_coefficient_1", ">i4"),
                                ("blackbody_temperature_2_conversion_coefficient_2", ">i4"),
                                ("blackbody_temperature_2_conversion_coefficient_3", ">i4"),
                                ("blackbody_temperature_2_conversion_coefficient_4", ">i4"),
                                ("blackbody_temperature_2_conversion_coefficient_5", ">i4"),
                                ("blackbody_temperature_2_conversion_coefficient_6", ">i4"),
                                ("blackbody_temperature_3_conversion_coefficient_1", ">i4"),
                                ("blackbody_temperature_3_conversion_coefficient_2", ">i4"),
                                ("blackbody_temperature_3_conversion_coefficient_3", ">i4"),
                                ("blackbody_temperature_3_conversion_coefficient_4", ">i4"),
                                ("blackbody_temperature_3_conversion_coefficient_5", ">i4"),
                                ("blackbody_temperature_3_conversion_coefficient_6", ">i4"),
                                ("blackbody_temperature_4_conversion_coefficient_1", ">i4"),
                                ("blackbody_temperature_4_conversion_coefficient_2", ">i4"),
                                ("blackbody_temperature_4_conversion_coefficient_3", ">i4"),
                                ("blackbody_temperature_4_conversion_coefficient_4", ">i4"),
                                ("blackbody_temperature_4_conversion_coefficient_5", ">i4"),
                                ("blackbody_temperature_4_conversion_coefficient_6", ">i4"),
                                ("electronics_current_conversion_coefficient_1", ">i4"),
                                ("electronics_current_conversion_coefficient_2", ">i4"),
                                ("electronics_current_conversion_coefficient_3", ">i4"),
                                ("electronics_current_conversion_coefficient_4", ">i4"),
                                ("electronics_current_conversion_coefficient_5", ">i4"),
                                ("electronics_current_conversion_coefficient_6", ">i4"),
                                ("motor_current_conversion_coefficient_1", ">i4"),
                                ("motor_current_conversion_coefficient_2", ">i4"),
                                ("motor_current_conversion_coefficient_3", ">i4"),
                                ("motor_current_conversion_coefficient_4", ">i4"),
                                ("motor_current_conversion_coefficient_5", ">i4"),
                                ("motor_current_conversion_coefficient_6", ">i4"),
                                ("earth_shield_position_conversion_coefficient_1", ">i4"),
                                ("earth_shield_position_conversion_coefficient_2", ">i4"),
                                ("earth_shield_position_conversion_coefficient_3", ">i4"),
                                ("earth_shield_position_conversion_coefficient_4", ">i4"),
                                ("earth_shield_position_conversion_coefficient_5", ">i4"),
                                ("earth_shield_position_conversion_coefficient_6", ">i4"),
                                ("electronics_temperature_conversion_coefficient_1", ">i4"),
                                ("electronics_temperature_conversion_coefficient_2", ">i4"),
                                ("electronics_temperature_conversion_coefficient_3", ">i4"),
                                ("electronics_temperature_conversion_coefficient_4", ">i4"),
                                ("electronics_temperature_conversion_coefficient_5", ">i4"),
                                ("electronics_temperature_conversion_coefficient_6", ">i4"),
                                ("cooler_housing_temperature_conversion_coefficient_1", ">i4"),
                                ("cooler_housing_temperature_conversion_coefficient_2", ">i4"),
                                ("cooler_housing_temperature_conversion_coefficient_3", ">i4"),
                                ("cooler_housing_temperature_conversion_coefficient_4", ">i4"),
                                ("cooler_housing_temperature_conversion_coefficient_5", ">i4"),
                                ("cooler_housing_temperature_conversion_coefficient_6", ">i4"),
                                ("baseplate_temperature_conversion_coefficient_1", ">i4"),
                                ("baseplate_temperature_conversion_coefficient_2", ">i4"),
                                ("baseplate_temperature_conversion_coefficient_3", ">i4"),
                                ("baseplate_temperature_conversion_coefficient_4", ">i4"),
                                ("baseplate_temperature_conversion_coefficient_5", ">i4"),
                                ("baseplate_temperature_conversion_coefficient_6", ">i4"),
                                ("motor_housing_temperature_conversion_coefficient_1", ">i4"),
                                ("motor_housing_temperature_conversion_coefficient_2", ">i4"),
                                ("motor_housing_temperature_conversion_coefficient_3", ">i4"),
                                ("motor_housing_temperature_conversion_coefficient_4", ">i4"),
                                ("motor_housing_temperature_conversion_coefficient_5", ">i4"),
                                ("motor_housing_temperature_conversion_coefficient_6", ">i4"),
                                ("a/d_converter_temperature_conversion_coefficient_1", ">i4"),
                                ("a/d_converter_temperature_conversion_coefficient_2", ">i4"),
                                ("a/d_converter_temperature_conversion_coefficient_3", ">i4"),
                                ("a/d_converter_temperature_conversion_coefficient_4", ">i4"),
                                ("a/d_converter_temperature_conversion_coefficient_5", ">i4"),
                                ("a/d_converter_temperature_conversion_coefficient_6", ">i4"),
                                ("detector_#4_bias_voltage_conversion_coefficient_1", ">i4"),
                                ("detector_#4_bias_voltage_conversion_coefficient_2", ">i4"),
                                ("detector_#4_bias_voltage_conversion_coefficient_3", ">i4"),
                                ("detector_#4_bias_voltage_conversion_coefficient_4", ">i4"),
                                ("detector_#4_bias_voltage_conversion_coefficient_5", ">i4"),
                                ("detector_#4_bias_voltage_conversion_coefficient_6", ">i4"),
                                ("detector_#5_bias_voltage_conversion_coefficient_1", ">i4"),
                                ("detector_#5_bias_voltage_conversion_coefficient_2", ">i4"),
                                ("detector_#5_bias_voltage_conversion_coefficient_3", ">i4"),
                                ("detector_#5_bias_voltage_conversion_coefficient_4", ">i4"),
                                ("detector_#5_bias_voltage_conversion_coefficient_5", ">i4"),
                                ("detector_#5_bias_voltage_conversion_coefficient_6", ">i4"),
                                ("blackbody_temperature_channel3b_conversion_coefficient_1", ">i4"),
                                ("blackbody_temperature_channel3b_conversion_coefficient_2", ">i4"),
                                ("blackbody_temperature_channel3b_conversion_coefficient_3", ">i4"),
                                ("blackbody_temperature_channel3b_conversion_coefficient_4", ">i4"),
                                ("blackbody_temperature_channel3b_conversion_coefficient_5", ">i4"),
                                ("blackbody_temperature_channel3b_conversion_coefficient_6", ">i4"),
                                ("blackbody_temperature_channel4_conversion_coefficient_1",
                                 ">i4"),
                                ("blackbody_temperature_channel4_conversion_coefficient_2",
                                 ">i4"),
                                ("blackbody_temperature_channel4_conversion_coefficient_3",
                                 ">i4"),
                                ("blackbody_temperature_channel4_conversion_coefficient_4",
                                 ">i4"),
                                ("blackbody_temperature_channel4_conversion_coefficient_5",
                                 ">i4"),
                                ("blackbody_temperature_channel4_conversion_coefficient_6",
                                 ">i4"),
                                ("blackbody_temperature_channel5_conversion_coefficient_1",
                                 ">i4"),
                                ("blackbody_temperature_channel5_conversion_coefficient_2",
                                 ">i4"),
                                ("blackbody_temperature_channel5_conversion_coefficient_3",
                                 ">i4"),
                                ("blackbody_temperature_channel5_conversion_coefficient_4",
                                 ">i4"),
                                ("blackbody_temperature_channel5_conversion_coefficient_5",
                                 ">i4"),
                                ("blackbody_temperature_channel5_conversion_coefficient_6",
                                 ">i4"),
                                ("reference_voltage_conversion_coefficient_1", ">i4"),
                                ("reference_voltage_conversion_coefficient_2", ">i4"),
                                ("reference_voltage_conversion_coefficient_3", ">i4"),
                                ("reference_voltage_conversion_coefficient_4", ">i4"),
                                ("reference_voltage_conversion_coefficient_5", ">i4"),
                                ("reference_voltage_conversion_coefficient_6", ">i4"),
                                # METOP MANEUVERS IDENTIFICATION
                                ("start_of_maneuver_year", ">u2"),
                                ("start_of_maneuver_day", ">u2"),
                                ("start_of_maneuver_utc_time_of_day", ">u4"),
                                ("end_of_manuever_year", ">u2"),
                                ("end_of_maneuver_day", ">u2"),
                                ("end_of_maneuver_utc_time_of_day", ">u4"),
                                ("change_in_spacecraft_velocity", ">i4"),
                                ("change in spacecraft_mass", ">u4"),
                                # CLOUDS FROM AVHRR(CLAVR)
                                ("clavr_status_bit_field", ">u2"),
                                ("zero_fill9", ">i2")])


ars_header = np.dtype([("COST_number", "S6"),
                       ("SAA_number", "S8"),
                       ("order_creation_year", "S4"),
                       ("order_creation_day_of_year", "S3"),
                       ("processing_site_code", "S1"),
                       ("processing_software", "S8"),
                       # data selection criteria
                       ("data_set_name", "S42"),
                       ("ascii_blank_", "S2"),
                       ("select_flag", "S1"),
                       ("beginning_latitude", "S3"),
                       ("ending_latitude", "S3"),
                       ("beginning_longitude", "S4"),
                       ("ending_longitude", "S4"),
                       ("start_hour", "S2"),
                       ("start_minute", "S2"),
                       ("number_of_minutes", "S3"),
                       ("appended_data_flag", "S1"),
                       ("channel_select_flag", "S1", (20, )),
                       # dataset summary
                       ("ascii_blank__", "S29"),
                       ("ascend_descend_flag", "S1"),
                       ("first_latitude", "S3"),
                       ("last_latitude", "S3"),
                       ("first_longitude", "S4"),
                       ("last_longitude", "S4"),
                       ("data_format", "S20"),
                       ("size_of_record", "S6"),
                       ("number_of_records", "S6"),
                       # filler
                       ("ascii_blank", "S319")
                       ])


class KLMReader(Reader):
    """Reader for KLM data."""

    spacecraft_names = {4: "noaa15",
                        2: "noaa16",
                        6: "noaa17",
                        7: "noaa18",
                        8: "noaa19",
                        12: "metopa",
                        11: "metopb",
                        13: "metopc",
                        }
    spacecrafts_orbital = {4: "noaa 15",
                           2: "noaa 16",
                           6: "noaa 17",
                           7: "noaa 18",
                           8: "noaa 19",
                           12: "metop 02",
                           11: "metop 01",
                           13: "metop 03",
                           }

    tsm_affected_intervals = TSM_AFFECTED_INTERVALS_KLM

    QFlag = KLM_QualityIndicator
    _quality_indicators_key = "quality_indicator_bit_field"

    def read(self, filename, fileobj=None):
        """Read the data.

        Args:
            filename: Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)

        Returns:
            header: numpy record array
                The header metadata
            scans: numpy record array
                The scanlines

        """
        # Note that np.fromfile does not work with gzip.GzipFile
        # objects (numpy version 1.16.4), because it restricts the
        # file objects to (io.FileIO, io.BufferedReader, io.BufferedWriter)
        # see: numpy.compat.py3k.isfileobj
        self.filename = filename
        LOG.info("Reading %s", self.filename)
        with file_opener(fileobj or filename) as fd_:
            self.ars_head, self.head = self.read_header(
                filename, fileobj=fd_)
            if self.ars_head:
                ars_offset = ars_header.itemsize
            else:
                ars_offset = 0
            self.header_version = self.head[
                "noaa_level_1b_format_version_number"]
            if self.header_version >= 5:
                self.analog_telemetry, = np.frombuffer(
                    fd_.read(analog_telemetry_v5.itemsize),
                    dtype=analog_telemetry_v5, count=1)
            else:
                self.analog_telemetry, = np.frombuffer(
                    fd_.read(analog_telemetry_v2.itemsize),
                    dtype=analog_telemetry_v2, count=1)
            # LAC: 1, GAC: 2, ...
            self.data_type = self.head["data_type_code"]
            # read until end of file
            fd_.seek(self.offset + ars_offset, 0)
            buffer = fd_.read()
            count = self.head["count_of_data_records"]
            self._read_scanlines(buffer, count)
        if self.correct_scanlines:
            self.correct_scan_line_numbers()
        self.spacecraft_id = self.head["noaa_spacecraft_identification_code"]
        self.spacecraft_name = self.spacecraft_names[self.spacecraft_id]

    @classmethod
    def read_header(cls, filename, fileobj=None):
        """Read the file header.

        Args:
            filename (str): Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)

        Returns:
            archive_header (struct): archive header
            header (struct): file header
        """
        with file_opener(fileobj or filename) as fd_:
            # read ars_header if present
            _ars_head, = np.frombuffer(
                fd_.read(ars_header.itemsize),
                dtype=ars_header, count=1)
            if _ars_head["data_format"].startswith(b"NOAA Level 1b"):
                ars_head = _ars_head.copy()
            else:
                fd_.seek(0)
                ars_head = None
            # need to copy frombuffer to have write access on head
            head, = np.frombuffer(
                fd_.read(header.itemsize),
                dtype=header, count=1).copy()
        head = cls._correct_data_set_name(head, filename)
        cls._validate_header(head)
        return ars_head, head

    @classmethod
    def _validate_header(cls, header):
        """Check if the header belongs to this reader."""
        # call super to enter the Method Resolution Order (MRO)
        super(KLMReader, cls)._validate_header(header)
        LOG.debug("validate header")
        data_set_name = header["data_set_name"].decode()
        # split header into parts
        creation_site, transfer_mode, platform_id = (
            data_set_name.split(".")[:3])
        allowed_ids = ["NK", "NL", "NM", "NN", "NP", "M1", "M2", "M3"]
        if platform_id not in allowed_ids:
            raise ReaderError('Improper platform id "%s"!' % platform_id)

    def get_telemetry(self):
        """Get the telemetry.

        Returns:
            prt_counts: np.array
            ict_counts: np.array
            space_counts: np.array

        """
        prt_counts = np.mean(self.scans["telemetry"]["PRT"], axis=1)

        # getting ICT counts

        ict_counts = np.zeros((len(self.scans), 3))
        ict_counts[:, 0] = np.mean(self.scans["back_scan"][:, 0::3], axis=1)
        ict_counts[:, 1] = np.mean(self.scans["back_scan"][:, 1::3], axis=1)
        ict_counts[:, 2] = np.mean(self.scans["back_scan"][:, 2::3], axis=1)

        # getting space counts

        space_counts = np.zeros((len(self.scans), 3))
        space_counts[:, 0] = np.mean(self.scans["space_data"][:, 2::5], axis=1)
        space_counts[:, 1] = np.mean(self.scans["space_data"][:, 3::5], axis=1)
        space_counts[:, 2] = np.mean(self.scans["space_data"][:, 4::5], axis=1)

        return prt_counts, ict_counts, space_counts

    def _get_lonlat_from_file(self):
        """Get the longitudes and latitudes."""
        lats = self.scans["earth_location"]["lats"] / np.float32(1e4)
        lons = self.scans["earth_location"]["lons"] / np.float32(1e4)
        return lons, lats

    def get_header_timestamp(self):
        """Get the timestamp from the header.

        Returns:
            A datetime object containing the timestamp from the header.

        Raises:
            A ValueError if the timestamp is corrupt.

        """
        year = self.head["start_of_data_set_year"]
        jday = self.head["start_of_data_set_day_of_year"]
        msec = self.head["start_of_data_set_utc_time_of_day"]
        try:
            return self.to_datetime(self.to_datetime64(year=year, jday=jday,
                                                       msec=msec))
        except ValueError as err:
            raise ValueError("Corrupt header timestamp: {0}".format(err))

    def _get_times_from_file(self):
        """Get the times of the scanlines."""
        year = self.scans["scan_line_year"]
        jday = self.scans["scan_line_day_of_year"]
        msec = self.scans["scan_line_utc_time_of_day"]
        return year, jday, msec

    def get_ch3_switch(self):
        """Channel 3 identification.

        0: Channel 3b (Brightness temperature)
        1: Channel 3a (Reflectance)
        2: Transition (No data)
        """
        return self.scans["scan_line_bit_field"][:] & 3

    def _get_ir_channels_to_calibrate(self):
        ir_channels_to_calibrate = [3, 4, 5]
        if np.all(self.get_ch3_switch() != 0):
            ir_channels_to_calibrate = [4, 5]
        return ir_channels_to_calibrate

    def postproc(self, ds):
        """Apply KLM specific postprocessing.

        Masking out 3a/3b/transition.
        """
        switch = self.get_ch3_switch()
        ds["channels"].loc[dict(channel_name="3a", scan_line_index=((switch==0) | (switch==2)))] = np.nan
        ds["channels"].loc[dict(channel_name="3b", scan_line_index=((switch==1) | (switch==2)))] = np.nan


    def _adjust_clock_drift(self):
        """Adjust the geolocation to compensate for the clock error.

        Note:
            Clock drift correction is only applied to POD satellites.
            On the KLM series, the clock is updated daily.
        """

    def get_tsm_pixels(self, channels):
        """Determine pixels affected by the scan motor issue.

        Uses channels 1, 2, 4 and 5. Neither 3a, nor 3b.
        """
        return get_tsm_idx(channels[:, :, 0], channels[:, :, 1],
                           channels[:, :, 4], channels[:, :, 5])


def main_klm(reader_cls, filename, start_line, end_line):
    """Generate a l1c file."""
    tic = datetime.datetime.now()
    reader = reader_cls.fromfile(filename)
    reader.save(int(start_line), int(end_line))
    LOG.info("pygac took: %s", str(datetime.datetime.now() - tic))



"""The LAC reader."""

import logging

import numpy as np

try:
    from pyorbital.geoloc_instrument_definitions import avhrr_from_times
except ImportError:
    # In pyorbital 1.9.2 and earlier avhrr_gac returned LAC geometry
    from pyorbital.geoloc_instrument_definitions import avhrr_gac as avhrr_from_times

from pygac.reader import Reader, ReaderError

LOG = logging.getLogger(__name__)


class LACReader(Reader):
    """Reader for LAC data."""

    # Scanning frequency (scanlines per millisecond)
    scan_freq = 6.0 / 1000.0
    # Max scanlines
    max_scanlines = 65535
    lonlat_sample_points = np.arange(24, 2048, 40)

    def __init__(self, *args, **kwargs):
        """Init the LAC reader."""
        super(LACReader, self).__init__(*args, **kwargs)
        self.scan_width = 2048
        self.geoloc_definition = avhrr_from_times

    @classmethod
    def _validate_header(cls, header):
        """Check if the header belongs to this reader."""
        # call super to enter the Method Resolution Order (MRO)
        super(LACReader, cls)._validate_header(header)
        LOG.debug("validate header")
        data_set_name = header["data_set_name"].decode()
        # split header into parts
        creation_site, transfer_mode, platform_id = (
            data_set_name.split(".")[:3])
        if transfer_mode not in ["LHRR", "HRPT", "FRAC"]:
            raise ReaderError('Improper transfer mode "%s"!' % transfer_mode)



import calendar
import datetime
import logging
import os
import time

import h5py
import numpy as np

from pygac.utils import check_user_scanlines, slice_channel, strip_invalid_lat

LOG = logging.getLogger(__name__)


MISSING_DATA = -32001
MISSING_DATA_LATLON = -999999


def save_gac(satellite_name,
             xutcs,
             lats, lons,
             ref1, ref2, ref3,
             bt3, bt4, bt5,
             sun_zen, sat_zen, sun_azi, sat_azi, rel_azi,
             qual_flags, start_line, end_line,
             gac_file, meta_data,
             output_file_prefix, avhrr_dir, qual_dir, sunsatangles_dir):

    midnight_scanline = meta_data["midnight_scanline"]
    miss_lines = meta_data["missing_scanlines"]
    corr = meta_data["sun_earth_distance_correction_factor"]

    last_scan_line_number = qual_flags[-1, 0]

    # Strip invalid coordinates
    first_valid_lat, last_valid_lat = strip_invalid_lat(lats)
    if first_valid_lat > start_line:
        LOG.info("New start_line chosen (due to invalid lat/lon "
                 "info) = " + str(first_valid_lat))
    if end_line > last_valid_lat:
        LOG.info("New end_line chosen (due to invalid lat/lon "
                 "info) = " + str(last_valid_lat))

    # Check user-defined scanlines
    start_line, end_line = check_user_scanlines(
        start_line=start_line,
        end_line=end_line,
        first_valid_lat=first_valid_lat,
        last_valid_lat=last_valid_lat)

    # Slice data using new start/end lines
    _, miss_lines, midnight_scanline = slice_channel(
        np.zeros(lats.shape),
        start_line=start_line,
        end_line=end_line,
        first_valid_lat=first_valid_lat,
        last_valid_lat=last_valid_lat,
        qual_flags=qual_flags,
        miss_lines=miss_lines,
        midnight_scanline=midnight_scanline)

    ref1, _, _ = slice_channel(ref1,
                               start_line=start_line,
                               end_line=end_line,
                               first_valid_lat=first_valid_lat,
                               last_valid_lat=last_valid_lat)
    ref2, _, _ = slice_channel(ref2,
                               start_line=start_line,
                               end_line=end_line,
                               first_valid_lat=first_valid_lat,
                               last_valid_lat=last_valid_lat)
    ref3, _, _ = slice_channel(ref3,
                               start_line=start_line,
                               end_line=end_line,
                               first_valid_lat=first_valid_lat,
                               last_valid_lat=last_valid_lat)
    bt3, _, _ = slice_channel(bt3,
                              start_line=start_line,
                              end_line=end_line,
                              first_valid_lat=first_valid_lat,
                              last_valid_lat=last_valid_lat)
    bt4, _, _ = slice_channel(bt4,
                              start_line=start_line,
                              end_line=end_line,
                              first_valid_lat=first_valid_lat,
                              last_valid_lat=last_valid_lat)
    bt5, _, _ = slice_channel(bt5,
                              start_line=start_line,
                              end_line=end_line,
                              first_valid_lat=first_valid_lat,
                              last_valid_lat=last_valid_lat)
    sun_zen, _, _ = slice_channel(sun_zen,
                                  start_line=start_line,
                                  end_line=end_line,
                                  first_valid_lat=first_valid_lat,
                                  last_valid_lat=last_valid_lat)
    sun_azi, _, _ = slice_channel(sun_azi,
                                  start_line=start_line,
                                  end_line=end_line,
                                  first_valid_lat=first_valid_lat,
                                  last_valid_lat=last_valid_lat)
    sat_zen, _, _ = slice_channel(sat_zen,
                                  start_line=start_line,
                                  end_line=end_line,
                                  first_valid_lat=first_valid_lat,
                                  last_valid_lat=last_valid_lat)
    sat_azi, _, _ = slice_channel(sat_azi,
                                  start_line=start_line,
                                  end_line=end_line,
                                  first_valid_lat=first_valid_lat,
                                  last_valid_lat=last_valid_lat)
    rel_azi, _, _ = slice_channel(rel_azi,
                                  start_line=start_line,
                                  end_line=end_line,
                                  first_valid_lat=first_valid_lat,
                                  last_valid_lat=last_valid_lat)
    lons, _, _ = slice_channel(lons,
                               start_line=start_line,
                               end_line=end_line,
                               first_valid_lat=first_valid_lat,
                               last_valid_lat=last_valid_lat)
    lats, _, _ = slice_channel(lats,
                               start_line=start_line,
                               end_line=end_line,
                               first_valid_lat=first_valid_lat,
                               last_valid_lat=last_valid_lat)
    qual_flags, _, _ = slice_channel(qual_flags,
                                     start_line=start_line,
                                     end_line=end_line,
                                     first_valid_lat=first_valid_lat,
                                     last_valid_lat=last_valid_lat)
    xutcs, _, _ = slice_channel(xutcs,
                                start_line=start_line,
                                end_line=end_line,
                                first_valid_lat=first_valid_lat,
                                last_valid_lat=last_valid_lat)

    total_number_of_scan_lines = lats.shape[0]

    # Reading time from the body of the gac file
    start = xutcs[0].astype(datetime.datetime)
    end = xutcs[-1].astype(datetime.datetime)
    startdate = start.strftime("%Y%m%d")
    starttime = start.strftime("%H%M%S%f")[:-5]
    enddate = end.strftime("%Y%m%d")
    endtime = end.strftime("%H%M%S%f")[:-5]

    # Apply scaling & offset
    bt3 -= 273.15
    bt4 -= 273.15
    bt5 -= 273.15
    for array in [bt3, bt4, bt5, ref1, ref2, ref3, sun_zen, sat_zen, sun_azi,
                  sat_azi, rel_azi]:
        array *= 100.0
    for array in [lats, lons]:
        array *= 1000.0

    # Replace NaN with fill values
    for array in [ref1, ref2, ref3, bt3, bt4, bt5, sun_zen, sat_zen, sun_azi,
                  sat_azi, rel_azi]:
        array[np.isnan(array)] = MISSING_DATA
    for array in [lats, lons]:
        array[np.isnan(array)] = MISSING_DATA_LATLON

    avhrrGAC_io(satellite_name, xutcs, startdate, enddate, starttime, endtime,
                lats, lons, ref1, ref2, ref3, bt3, bt4, bt5,
                sun_zen, sat_zen, sun_azi, sat_azi, rel_azi, qual_flags,
                start_line, end_line, total_number_of_scan_lines,
                last_scan_line_number, corr, gac_file, midnight_scanline,
                miss_lines, output_file_prefix, avhrr_dir, qual_dir, sunsatangles_dir)


def avhrrGAC_io(satellite_name, xutcs, startdate, enddate, starttime, endtime,
                arrLat_full, arrLon_full, ref1, ref2, ref3, bt3, bt4, bt5,
                arrSZA, arrSTZ, arrSAA, arrSTA, arrRAA, qual_flags,
                start_line, end_line, total_number_of_scan_lines,
                last_scan_line_number, corr, gac_file, midnight_scanline,
                miss_lines, output_file_prefix, avhrr_dir, qual_dir, sunsatangles_dir):

    # Calculate start and end time in sec1970
    t_obj = time.strptime(startdate + starttime[0:6], "%Y%m%d%H%M%S")
    starttime_sec1970 = calendar.timegm(t_obj)
    t_obj = time.strptime(enddate + endtime[0:6], "%Y%m%d%H%M%S")
    endtime_sec1970 = calendar.timegm(t_obj)

    LOG.info("Output file prefix = " + str(output_file_prefix))
    LOG.info("AVHRR data will be written to " + str(avhrr_dir))
    ofn = os.path.join(avhrr_dir, (output_file_prefix + "_avhrr_" +
                                   satellite_name + "_99999_" +
                                   startdate + "T" + starttime + "Z_" +
                                   enddate + "T" + endtime + "Z.h5"))

    LOG.info("Filename: " + str(os.path.basename(ofn)))

    fout = h5py.File(ofn, "w")

    dset1 = fout.create_dataset("/image1/data", dtype="int16", data=ref1)
    dset2 = fout.create_dataset("/image2/data", dtype="int16", data=ref2)
    dset3 = fout.create_dataset("/image3/data", dtype="int16", data=bt3)
    dset4 = fout.create_dataset("/image4/data", dtype="int16", data=bt4)
    dset5 = fout.create_dataset("/image5/data", dtype="int16", data=bt5)
    dset6 = fout.create_dataset("/image6/data", dtype="int16", data=ref3)
    dset7 = fout.create_dataset("/where/lat/data", dtype="int32",
                                data=arrLat_full)
    dset8 = fout.create_dataset("/where/lon/data", dtype="int32",
                                data=arrLon_full)
    del dset8
    channellist = []
    channellist.append("channel1".encode("utf8"))
    channellist.append("channel2".encode("utf8"))
    channellist.append("channel3b".encode("utf8"))
    channellist.append("channel4".encode("utf8"))
    channellist.append("channel5".encode("utf8"))
    channellist.append("channel3a".encode("utf8"))
    dset10 = fout.create_dataset("/how/channel_list",
                                 data=channellist)
    del dset10
    # Attributes directly on highest level groups
    g1 = fout.require_group("/image1")
    g2 = fout.require_group("/image2")
    g3 = fout.require_group("/image3")
    g4 = fout.require_group("/image4")
    g5 = fout.require_group("/image5")
    g6 = fout.require_group("/image6")
    g7 = fout.require_group("/where")

    g1.attrs["channel"] = np.bytes_("1")
    g1.attrs["description"] = np.bytes_("AVHRR ch1")
    g2.attrs["channel"] = np.bytes_("2")
    g2.attrs["description"] = np.bytes_("AVHRR ch2")
    g3.attrs["channel"] = np.bytes_("3b")
    g3.attrs["description"] = np.bytes_("AVHRR ch3b")
    g4.attrs["channel"] = np.bytes_("4")
    g4.attrs["description"] = np.bytes_("AVHRR ch4")
    g5.attrs["channel"] = np.bytes_("5")
    g5.attrs["description"] = np.bytes_("AVHRR ch5")
    g6.attrs["channel"] = np.bytes_("3a")
    g6.attrs["description"] = np.bytes_("AVHRR ch3a")
    g7.attrs["num_of_pixels"] = np.int32(arrSZA.shape[1])
    g7.attrs["num_of_lines"] = np.int32(arrSZA.shape[0])
    g7.attrs["xscale"] = np.float32(0.0)
    g7.attrs["yscale"] = np.float32(0.0)
    g7.attrs["start_line"] = start_line
    g7.attrs["end_line"] = end_line

    # Attributes in the 'what' groups
    g1 = fout.create_group("/image1/what")
    g2 = fout.create_group("/image2/what")
    g3 = fout.create_group("/image3/what")
    g4 = fout.create_group("/image4/what")
    g5 = fout.create_group("/image5/what")
    g6 = fout.create_group("/image6/what")
    g7 = fout.create_group("/where/lat/what")
    g8 = fout.create_group("/where/lon/what")
    g9 = fout.create_group("/what")

    g1.attrs["product"] = np.bytes_("SATCH")
    g1.attrs["quantity"] = np.bytes_("REFL")
    g1.attrs["dataset_name"] = np.bytes_("Channel 1 reflectance")
    g1.attrs["units"] = np.bytes_("%")
    g1.attrs["gain"] = np.float32(0.01)
    g1.attrs["offset"] = np.float32(0.0)
    g1.attrs["missingdata"] = np.int32(MISSING_DATA)
    g1.attrs["nodata"] = np.int32(MISSING_DATA)
    g1.attrs["starttime"] = np.bytes_(starttime[0:6])
    g1.attrs["endtime"] = np.bytes_(endtime[0:6])
    g1.attrs["startdate"] = np.bytes_(startdate)
    g1.attrs["enddate"] = np.bytes_(enddate)

    g2.attrs["product"] = np.bytes_("SATCH")
    g2.attrs["quantity"] = np.bytes_("REFL")
    g2.attrs["dataset_name"] = np.bytes_("Channel 2 reflectance")
    g2.attrs["units"] = np.bytes_("%")
    g2.attrs["gain"] = np.float32(0.01)
    g2.attrs["offset"] = np.float32(0.0)
    g2.attrs["missingdata"] = np.int32(MISSING_DATA)
    g2.attrs["nodata"] = np.int32(MISSING_DATA)
    g2.attrs["starttime"] = np.bytes_(starttime[0:6])
    g2.attrs["endtime"] = np.bytes_(endtime[0:6])
    g2.attrs["startdate"] = np.bytes_(startdate)
    g2.attrs["enddate"] = np.bytes_(enddate)

    g6.attrs["product"] = np.bytes_("SATCH")
    g6.attrs["quantity"] = np.bytes_("REFL")
    g6.attrs["dataset_name"] = np.bytes_("Channel 3a reflectance")
    g6.attrs["units"] = np.bytes_("%")
    g6.attrs["gain"] = np.float32(0.01)
    g6.attrs["offset"] = np.float32(0.0)
    g6.attrs["missingdata"] = np.int32(MISSING_DATA)
    g6.attrs["nodata"] = np.int32(MISSING_DATA)
    g6.attrs["starttime"] = np.bytes_(starttime[0:6])
    g6.attrs["endtime"] = np.bytes_(endtime[0:6])
    g6.attrs["startdate"] = np.bytes_(startdate)
    g6.attrs["enddate"] = np.bytes_(enddate)

    g3.attrs["product"] = np.bytes_("SATCH")
    g3.attrs["quantity"] = np.bytes_("TB")
    g3.attrs["dataset_name"] = np.bytes_("Channel 3b brightness temperature")
    g3.attrs["units"] = np.bytes_("K")
    g3.attrs["gain"] = np.float32(0.01)
    g3.attrs["offset"] = np.float32(273.15)
    g3.attrs["missingdata"] = np.int32(MISSING_DATA)
    g3.attrs["nodata"] = np.int32(MISSING_DATA)
    g3.attrs["starttime"] = np.bytes_(starttime[0:6])
    g3.attrs["endtime"] = np.bytes_(endtime[0:6])
    g3.attrs["startdate"] = np.bytes_(startdate)
    g3.attrs["enddate"] = np.bytes_(enddate)

    g4.attrs["product"] = np.bytes_("SATCH")
    g4.attrs["quantity"] = np.bytes_("TB")
    g4.attrs["dataset_name"] = np.bytes_("Channel 4 brightness temperature")
    g4.attrs["units"] = np.bytes_("K")
    g4.attrs["gain"] = np.float32(0.01)
    g4.attrs["offset"] = np.float32(273.15)
    g4.attrs["missingdata"] = np.int32(MISSING_DATA)
    g4.attrs["nodata"] = np.int32(MISSING_DATA)
    g4.attrs["starttime"] = np.bytes_(starttime[0:6])
    g4.attrs["endtime"] = np.bytes_(endtime[0:6])
    g4.attrs["startdate"] = np.bytes_(startdate)
    g4.attrs["enddate"] = np.bytes_(enddate)

    g5.attrs["product"] = np.bytes_("SATCH")
    g5.attrs["quantity"] = np.bytes_("TB")
    g5.attrs["dataset_name"] = np.bytes_("Channel 5 brightness temperature")
    g5.attrs["units"] = np.bytes_("K")
    g5.attrs["gain"] = np.float32(0.01)
    g5.attrs["offset"] = np.float32(273.15)
    g5.attrs["missingdata"] = np.int32(MISSING_DATA)
    g5.attrs["nodata"] = np.int32(MISSING_DATA)
    g5.attrs["starttime"] = np.bytes_(starttime[0:6])
    g5.attrs["endtime"] = np.bytes_(endtime[0:6])
    g5.attrs["startdate"] = np.bytes_(startdate)
    g5.attrs["enddate"] = np.bytes_(enddate)

    g7.attrs["dataset_name"] = np.bytes_("Latitude")
    g7.attrs["units"] = np.bytes_("Deg")
    g7.attrs["gain"] = np.float32(0.0010)
    g7.attrs["offset"] = np.float32(0.0)
    g7.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["starttime"] = np.bytes_(starttime[0:6])
    g7.attrs["endtime"] = np.bytes_(endtime[0:6])
    g7.attrs["startdate"] = np.bytes_(startdate)
    g7.attrs["enddate"] = np.bytes_(enddate)

    g8.attrs["dataset_name"] = np.bytes_("Longitude")
    g8.attrs["units"] = np.bytes_("Deg")
    g8.attrs["gain"] = np.float32(0.0010)
    g8.attrs["offset"] = np.float32(0.0)
    g8.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g8.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g8.attrs["starttime"] = np.bytes_(starttime[0:6])
    g8.attrs["endtime"] = np.bytes_(endtime[0:6])
    g8.attrs["startdate"] = np.bytes_(startdate)
    g8.attrs["enddate"] = np.bytes_(enddate)

    g9.attrs["object"] = np.bytes_("SATP")
    g9.attrs["sets"] = np.int32(len(channellist))
    g9.attrs["version"] = np.bytes_("H5rad ?.?")
    g9.attrs["date"] = np.bytes_(startdate)
    g9.attrs["time"] = np.bytes_(starttime[0:6])

    # Attributes in the 'how' groups
    g1 = fout.create_group("/image1/how")
    g2 = fout.create_group("/image2/how")
    g3 = fout.create_group("/image3/how")
    g4 = fout.create_group("/image4/how")
    g5 = fout.create_group("/image5/how")
    g6 = fout.create_group("/image6/how")
    g10 = fout.require_group("/how")

    # SHq: Is the sun_earth_distance correction applied?
    g1.attrs["sun_earth_distance_correction_applied"] = np.bytes_("TRUE")
    g1.attrs["sun_earth_distance_correction_factor"] = corr
    g2.attrs["sun_earth_distance_correction_applied"] = np.bytes_("TRUE")
    g2.attrs["sun_earth_distance_correction_factor"] = corr
    # No attributes on 'how' for image3,4,5
    g6.attrs["sun_earth_distance_correction_applied"] = np.bytes_("TRUE")
    g6.attrs["sun_earth_distance_correction_factor"] = corr

    # We do not know much about how; mostly use no-data
    g10.attrs["yaw_error"] = 0.0
    g10.attrs["roll_error"] = 0.0
    g10.attrs["pitch_error"] = 0.0
    g10.attrs["startepochs"] = starttime_sec1970
    g10.attrs["endepochs"] = endtime_sec1970
    g10.attrs["platform"] = np.bytes_(satellite_name)
    g10.attrs["instrument"] = np.bytes_("avhrr")
    g10.attrs["orbit_number"] = np.int32(99999)
    g10.attrs["gac_file"] = np.bytes_(gac_file)
    g10.attrs["software"] = np.bytes_("pyGAC")
    g10.attrs["version"] = np.bytes_("1.0")

    fout.close()

    LOG.info("Sun and Satellite viewing angles will be " +
             "written to " + str(sunsatangles_dir))
    ofn = os.path.join(sunsatangles_dir,
                       (output_file_prefix + "_sunsatangles_" +
                        satellite_name + "_99999_" + startdate +
                        "T" + starttime + "Z_" +
                        enddate + "T" + endtime + "Z.h5"))

    LOG.info("Filename: " + str(os.path.basename(ofn)))
    fout = h5py.File(ofn, "w")

    dset1 = fout.create_dataset("/image1/data", dtype="int16", data=arrSZA)
    dset2 = fout.create_dataset("/image2/data", dtype="int16", data=arrSTZ)
    dset3 = fout.create_dataset("/image3/data", dtype="int16", data=arrRAA)
    dset4 = fout.create_dataset("/image4/data", dtype="int16", data=arrSAA)
    dset5 = fout.create_dataset("/image5/data", dtype="int16", data=arrSTA)
    dset6 = fout.create_dataset("/where/lat/data", dtype="int32",
                                data=arrLat_full)
    dset7 = fout.create_dataset("/where/lon/data", dtype="int32",
                                data=arrLon_full)

    del dset4, dset5, dset6, dset7

    # Attributes directly on highest level groups
    g1 = fout.require_group("/image1")
    g2 = fout.require_group("/image2")
    g3 = fout.require_group("/image3")
    g4 = fout.require_group("/image4")
    g5 = fout.require_group("/image5")
    g6 = fout.require_group("/where")

    g1.attrs["description"] = np.bytes_("Solar zenith angle")
    g2.attrs["description"] = np.bytes_("Satellite zenith angle")
    g3.attrs["description"] = np.bytes_(
        "Relative satellite-sun azimuth angle")
    g4.attrs["description"] = np.bytes_("Solar azimuth angle")
    g5.attrs["description"] = np.bytes_("Satellite azimuth angle")
    g6.attrs["num_of_pixels"] = np.int32(arrSZA.shape[1])
    g6.attrs["num_of_lines"] = np.int32(arrSZA.shape[0])
    g6.attrs["xscale"] = np.float32(0.0)
    g6.attrs["yscale"] = np.float32(0.0)
    g6.attrs["start_line"] = start_line
    g6.attrs["end_line"] = end_line

    # Attributes in the 'what' groups + 'how'
    g1 = fout.create_group("/image1/what")
    g2 = fout.create_group("/image2/what")
    g3 = fout.create_group("/image3/what")
    g4 = fout.create_group("/image4/what")
    g5 = fout.create_group("/image5/what")
    g6 = fout.create_group("/where/lat/what")
    g7 = fout.create_group("/where/lon/what")
    g8 = fout.create_group("/what")
    g9 = fout.create_group("/how")

    g1.attrs["product"] = np.bytes_("SUNZ")
    g1.attrs["quantity"] = np.bytes_("DEG")
    g1.attrs["dataset_name"] = np.bytes_("Solar zenith angle")
    g1.attrs["units"] = np.bytes_("Deg")
    g1.attrs["gain"] = np.float32(0.01)
    g1.attrs["offset"] = np.float32(0.0)
    g1.attrs["missingdata"] = np.int32(MISSING_DATA)
    g1.attrs["nodata"] = np.int32(MISSING_DATA)
    g1.attrs["starttime"] = np.bytes_(starttime[0:6])
    g1.attrs["endtime"] = np.bytes_(endtime[0:6])
    g1.attrs["startdate"] = np.bytes_(startdate)
    g1.attrs["enddate"] = np.bytes_(enddate)

    g2.attrs["product"] = np.bytes_("SATZ")
    g2.attrs["quantity"] = np.bytes_("DEG")
    g2.attrs["dataset_name"] = np.bytes_("Satellite zenith angle")
    g2.attrs["units"] = np.bytes_("Deg")
    g2.attrs["gain"] = np.float32(0.01)
    g2.attrs["offset"] = np.float32(0.0)
    g2.attrs["missingdata"] = np.int32(MISSING_DATA)
    g2.attrs["nodata"] = np.int32(MISSING_DATA)
    g2.attrs["starttime"] = np.bytes_(starttime[0:6])
    g2.attrs["endtime"] = np.bytes_(endtime[0:6])
    g2.attrs["startdate"] = np.bytes_(startdate)
    g2.attrs["enddate"] = np.bytes_(enddate)

    g3.attrs["product"] = np.bytes_("SSAZD")
    g3.attrs["quantity"] = np.bytes_("DEG")
    g3.attrs["dataset_name"] = np.bytes_(
        "Relative satellite-sun azimuth angle")
    g3.attrs["units"] = np.bytes_("Deg")
    g3.attrs["gain"] = np.float32(0.01)
    g3.attrs["offset"] = np.float32(0.0)
    g3.attrs["missingdata"] = np.int32(MISSING_DATA)
    g3.attrs["nodata"] = np.int32(MISSING_DATA)
    g3.attrs["starttime"] = np.bytes_(starttime[0:6])
    g3.attrs["endtime"] = np.bytes_(endtime[0:6])
    g3.attrs["startdate"] = np.bytes_(startdate)
    g3.attrs["enddate"] = np.bytes_(enddate)

    g4.attrs["product"] = np.bytes_("SUNA")
    g4.attrs["quantity"] = np.bytes_("DEG")
    g4.attrs["dataset_name"] = np.bytes_("Solar azimuth angle")
    g4.attrs["units"] = np.bytes_("Deg")
    g4.attrs["gain"] = np.float32(0.01)
    g4.attrs["offset"] = np.float32(0.0)
    g4.attrs["missingdata"] = np.int32(MISSING_DATA)
    g4.attrs["nodata"] = np.int32(MISSING_DATA)
    g4.attrs["starttime"] = np.bytes_(starttime[0:6])
    g4.attrs["endtime"] = np.bytes_(endtime[0:6])
    g4.attrs["startdate"] = np.bytes_(startdate)
    g4.attrs["enddate"] = np.bytes_(enddate)

    g5.attrs["product"] = np.bytes_("SATA")
    g5.attrs["quantity"] = np.bytes_("DEG")
    g5.attrs["dataset_name"] = np.bytes_("Satellite azimuth angle")
    g5.attrs["units"] = np.bytes_("Deg")
    g5.attrs["gain"] = np.float32(0.01)
    g5.attrs["offset"] = np.float32(0.0)
    g5.attrs["missingdata"] = np.int32(MISSING_DATA)
    g5.attrs["nodata"] = np.int32(MISSING_DATA)
    g5.attrs["starttime"] = np.bytes_(starttime[0:6])
    g5.attrs["endtime"] = np.bytes_(endtime[0:6])
    g5.attrs["startdate"] = np.bytes_(startdate)
    g5.attrs["enddate"] = np.bytes_(enddate)

    g6.attrs["dataset_name"] = np.bytes_("Latitude")
    g6.attrs["units"] = np.bytes_("Deg")
    g6.attrs["gain"] = np.float32(0.0010)
    g6.attrs["offset"] = np.float32(0.0)
    g6.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g6.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g6.attrs["starttime"] = np.bytes_(starttime[0:6])
    g6.attrs["endtime"] = np.bytes_(endtime[0:6])
    g6.attrs["startdate"] = np.bytes_(startdate)
    g6.attrs["enddate"] = np.bytes_(enddate)

    g7.attrs["dataset_name"] = np.bytes_("Longitude")
    g7.attrs["units"] = np.bytes_("Deg")
    g7.attrs["gain"] = np.float32(0.0010)
    g7.attrs["offset"] = np.float32(0.0)
    g7.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["starttime"] = np.bytes_(starttime[0:6])
    g7.attrs["endtime"] = np.bytes_(endtime[0:6])
    g7.attrs["startdate"] = np.bytes_(startdate)
    g7.attrs["enddate"] = np.bytes_(enddate)

    g8.attrs["object"] = np.bytes_("SATP")
    g8.attrs["sets"] = np.int32(5)
    g8.attrs["version"] = np.bytes_("H5rad ?.?")
    g8.attrs["date"] = np.bytes_(startdate)
    g8.attrs["time"] = np.bytes_(starttime[0:6])

    # We do not know much about how; mostly use no-data
    g9.attrs["yaw_error"] = 0.0
    g9.attrs["roll_error"] = 0.0
    g9.attrs["pitch_error"] = 0.0
    g9.attrs["startepochs"] = starttime_sec1970
    g9.attrs["endepochs"] = endtime_sec1970
    g9.attrs["platform"] = np.bytes_(satellite_name)
    g9.attrs["instrument"] = np.bytes_("avhrr")
    g9.attrs["orbit_number"] = np.int32(99999)
    g9.attrs["gac_file"] = np.bytes_(gac_file)
    g9.attrs["software"] = np.bytes_("pyGAC")
    g9.attrs["version"] = np.bytes_("1.0")

    fout.close()

    LOG.info("Quality flags will be " +
             "written to " + str(qual_dir))
    ofn = os.path.join(qual_dir,
                       (output_file_prefix + "_qualflags_" +
                        satellite_name + "_99999_" + startdate +
                        "T" + starttime + "Z_" +
                        enddate + "T" + endtime + "Z.h5"))

    LOG.info("Filename: " + str(os.path.basename(ofn)))
    fout = h5py.File(ofn, "w")

    g1 = fout.require_group("/qual_flags")
    dset1 = g1.create_dataset("data", dtype="int16", data=qual_flags)
    del dset1

    g1.attrs["product"] = np.bytes_("QFLAG")
    g1.attrs["quantity"] = np.bytes_("INT")
    g1.attrs["dataset_name"] = np.bytes_("Scanline quality flags")
    g1.attrs["units"] = np.bytes_("None")
    g1.attrs["gain"] = np.int32(1)
    g1.attrs["offset"] = np.int32(0)
    g1.attrs["missingdata"] = np.int32(MISSING_DATA)
    g1.attrs["nodata"] = np.int32(MISSING_DATA)
    g1.attrs["starttime"] = np.bytes_(starttime[0:6])
    g1.attrs["endtime"] = np.bytes_(endtime[0:6])
    g1.attrs["startdate"] = np.bytes_(startdate)
    g1.attrs["enddate"] = np.bytes_(enddate)
    g1.attrs["gac_file"] = np.bytes_(gac_file)
    g1.attrs["total_number_of_data_records"] = total_number_of_scan_lines
    g1.attrs["last_scan_line_number"] = last_scan_line_number

    g2 = fout.require_group("/ancillary")
    dset2 = g2.create_dataset("missing_scanlines", dtype="int16",
                              data=miss_lines)
    del dset2
    dset3 = g2.create_dataset("scanline_timestamps", dtype="int64",
                              data=xutcs.astype("int64"))
    dset3.attrs["units"] = "Milliseconds since 1970-01-01 00:00:00 UTC"
    dset3.attrs["calendar"] = "standard"
    g2.attrs["midnight_scanline"] = np.bytes_(midnight_scanline)

    fout.close()
